import unittest
from dspyce.bitstreams import Bundle
from dspyce.bitstreams import Bitstream


class BitstreamsTest(unittest.TestCase):
    bundle: Bundle = Bundle()
    bitstream: Bitstream = Bitstream('other', 'test-file')

    def test_init_bundle(self):
        self.assertIsInstance(self.bundle, Bundle)
        self.assertEqual(Bundle.DEFAULT_BUNDLE, self.bundle.name)

    def test_init_bitstream(self):
        self.assertIsInstance(self.bitstream, Bitstream)
        self.assertEqual('other', self.bitstream.file_name)

    def test_bundle_bitstream(self):
        self.bundle.add_bitstream(self.bitstream)
        self.assertIn(self.bitstream, self.bundle.get_bitstreams())
        self.bundle.remove_bitstream(self.bitstream)
        self.assertNotIn(self.bitstream, self.bundle.get_bitstreams())



from dspyce.bitstreams import Bitstream


class Bundle:
    """
    The class Bundle represents a bundle in the DSpace context. I can contain several bitstreams.
    """

    uuid: str | None
    """The uuid of the bundle"""
    DEFAULT_BUNDLE: str = 'ORIGINAL'
    """The default bundle name."""
    name: str
    """The bundle name."""
    description: str
    """A bundle description if existing."""
    bitstreams: list[Bitstream]
    """A list of bitstream associated with the bundle"""

    def __init__(self, name: str = DEFAULT_BUNDLE, description: str = '', uuid: str = None,
                 bitstreams: list[Bitstream] = None):
        """
        Creates a new bundle object.

        :param name: The bundle name.
        :param description: A description if existing.
        :param uuid: The uuid of the bundle, if known.
        :param bitstreams: A list of bitstreams associated with this bundle.
        :raises AttributeError: If the bundle name is not of type <str> or is empty.
        """
        if not isinstance(name, str) or name.strip() == '':
            raise AttributeError('You have to provide a correct bundle name expected not-empty string,'
                                 f'but got "{name}"')
        self.name = name
        self.uuid = uuid
        self.description = description
        self.bitstreams = []
        if bitstreams is not None:
            [self.add_bitstream(b) for b in bitstreams]

    def __str__(self):
        return ('Bundle - {}{}:\n{}'.format(self.name,
                                            f'({self.uuid})' if self.uuid is not None else '',
                                            '\n'.join([f'\t{i}' for i in self.bitstreams])))

    def __eq__(self, other) -> bool:
        """
        Check if two bundle objects are equal

        :param other: The other bundle object to compare with.
        :return: True, if the two bundles have the same name.
        """
        if not isinstance(other, Bundle):
            raise TypeError(f'Can not compare type Bundle to "{type(other)}"')
        if self.uuid is None or other.uuid is None:
            return self.name == other.name

        return self.uuid == other.uuid and self.name == other.name

    def get_bitstreams(self, filter_condition=lambda x: True) -> list[Bitstream]:
        """
        Returns a list of bitstreams in this bundle, filtered by a filter defined in filter_condition.

        :param filter_condition: A condition to filter the bitstreams returned.
        :return: A list of Bitstream objects.
        """
        return list(filter(filter_condition, self.bitstreams))

    def add_bitstream(self, bitstream: Bitstream):
        """
        Adds a bitstream to this bundle.

        :param bitstream: The bitstream to add.
        """
        bitstream.bundle = self
        self.bitstreams.append(bitstream)

    def remove_bitstream(self, bitstream: Bitstream):
        """
        Removes a bitstream from the current bundle. Raises an Exception if the bitstream does not exist.

        :param bitstream: The bitstream to remove from the bundle.
        :raises FileNotFoundError: If the bitstream does not exist.
        """
        self.bitstreams.remove(bitstream)

    def save_bitstreams(self, path: str):
        """
        Saves the bitstreams of the given bundle into path.

        :param path: The path where to save the bitstreams.
        :raises FileExistsError: If the Bitstream already exists in the given path.
        """
        for b in self.bitstreams:
            b.save_bitstream(path)



from getpass import getpass
from .RestAPI import RestAPI
import requests
import logging


def authenticate_to_rest(rest_api: str, user: str = None, log_level=logging.INFO, log_file: str = None) -> RestAPI:
    """
    Connect to a given REST-API and ask for username and password via commandline.

    :param rest_api: The url of the REST-API endpoint.
    :param user: A username, if already known. If None, the username will be retrieved from input().
    :param log_level: The log_level used for Logging. Must be string or integer. The strings must be one of
        the following: DEBUG, INFO, WARNING, ERROR, CRITICAL. Default is INFO.
    :param log_file: A possible name and path of the log file. If None provided, all output will be logged to the
        console.
    :return: An object of the class Rest.
    """
    print(
        f'Establishing connection to the REST-API "{rest_api}"' + (f' with user "{user}":' if user is not None else ':')
    )
    authentication = False
    rest_server = None
    while not authentication:
        username = input('\tPlease enter your username: ') if user is None else user
        password = getpass('\tPassword: ')
        try:
            rest_server = RestAPI(rest_api, username, password, log_level, log_file)
        except requests.exceptions.JSONDecodeError:
            print(f'Did not found the api-Endpoint. Are you sure, that {rest_api} is the correct address and the API is'
                  'reachable?')
        authentication = rest_server.authenticated
        if not authentication:
            print('Wrong username or password! Please try again.')
    return rest_server



from .DSpaceObject import DSpaceObject


class Community(DSpaceObject):
    """
    The class Community represents a DSpace community containing sub communities or collections.
    """
    parent_community: DSpaceObject | None
    sub_communities: list[DSpaceObject]

    def __init__(self, uuid: str = '', handle: str = '', name: str = '', parent_community: DSpaceObject = None,
                 sub_communities: list[DSpaceObject] = None):
        super().__init__(uuid, handle, name)
        self.sub_communities = sub_communities if sub_communities is not None else []
        self.parent_community = parent_community

    def is_subcommunity_of(self, other) -> bool:
        """
        Checks if the community object is a sub-community of the given community.
        :param other: Another community object.
        :return: True, if self is in other.sub_communities.
        """
        if not isinstance(other, Community):
            raise TypeError('The given object must be of the type "Community".')
        return self in other.sub_communities

    def get_dspace_object_type(self) -> str:
        return 'Community'



from dspyce.Collection import Collection
from dspyce.DSpaceObject import DSpaceObject
from dspyce.Relation import Relation
from dspyce.bitstreams import Bitstream, IIIFBitstream, Bundle


class Item(DSpaceObject):
    """
    The class Item represents a single DSpace item. It can have a owning collection, several Bitstreams or relations
    to other items, if it's an entity.
    """
    collections: list[Collection]
    relations: list[Relation]
    contents: list[Bitstream]
    bundles: list[Bundle]
    in_archive: bool = True
    discoverable: bool = True
    withdrawn: bool = False

    def __init__(self, uuid: str = '', handle: str = '', name: str = '',
                 collections: Collection | list[Collection] | str = None):
        """
        Creates a new object of the Item class.

        :param uuid: The uuid of the Item.
        :param handle: The handle of the Item.
        :param name: The name of the DSpace Item, if existing.
        :param collections: Collections connected to this item. The first collection in the list will be the owning
        collection. Just an uuid can also be provided.
        """
        super().__init__(uuid, handle, name)
        if isinstance(collections, str):
            self.collections = [Collection(uuid=collections, community=None)]
        elif isinstance(collections, Collection):
            self.collections = [collections]
        elif collections is None:
            self.collections = []
        else:
            self.collections = collections
        self.relations = []
        self.contents = []
        self.bundles = []

    def is_entity(self) -> bool:
        """
        Checks if the item is a DSpace-Entity (True, if the metadata field dspace.entity.type is not empty).

        :return: True, if the Item is an entity.
        """
        return self.metadata.get('dspace.entity.type') is not None

    def get_entity_type(self) -> str | None:
        """
        Checks if the item is a DSpace-Entity and returns the value of dspace.entity.type if true, if not it returns
        None.

        :return: The entity type as a string, if existing, else None.
        """
        if self.is_entity():
            return self.get_metadata_values('dspace.entity.type')[0]

        return None

    def add_collection(self, c: Collection, primary: bool = False):
        """
        Adds an owning collection to the item. If primary is True, the collection will be set as the owning collection.
        :param c:
        :param primary:
        :return:
        """
        if primary:
            self.collections = [c] + self.collections
        else:
            self.collections.append(c)

    def add_relation(self, relation_type: str, identifier: str):
        """
        Adds a new relation to another item. Only possible if `is_entity()==True`.

        :param relation_type: The name of the relationship
        :param identifier: The identifier of the related Item.
        :return: None
        """
        if not self.is_entity():
            raise TypeError('Could not add relations to a non entity item for item:\n' + str(self))
        self.relations.append(Relation(relation_type, (self, Item(uuid=identifier))))

    def add_content(self, content_file: str, path: str, description: str = '', bundle: str | Bundle = Bundle(),
                    permissions: list[tuple[str, str]] = None, iiif: bool = False, width: int = 0, iiif_toc: str = ''):
        """
        Adds additional content-files to the item.

        :param content_file: The name of the document, which should be added.
        :param path: The path where to find the document.
        :param description: A description of the content file.
        :param bundle: The bundle where the item is stored in. The default is bundle.DEFAULT_BUNDLE
        :param permissions: Add permissions to a content file. This variable expects a list of tuples containing the
            permission-type and the group name to which it is granted to.
        :param iiif: If the bitstream should be treated as an iiif-specific file. If true also "dspace.iiif.enabled"
            will be set to "true".
        :param width: The width of an image. Only needed, if the file is a jpg, wich should be reduced and iiif is True.
        :param iiif_toc: A toc information for an iiif-specific bitstream.
        """
        active_bundle = self.get_bundle(bundle.name if isinstance(bundle, Bundle) else bundle)
        if active_bundle is None:
            active_bundle = bundle if isinstance(bundle, Bundle) else Bundle(bundle)
            self.bundles.append(active_bundle)

        if iiif:
            cf = IIIFBitstream(content_file, path, bundle=bundle)
            name = content_file.split('.')[0]
            cf.add_iiif(description, name if iiif_toc == '' else iiif_toc, w=width)
            if self.metadata.get('dspace.iiif.enabled') is None:
                self.add_metadata('dspace.iiif.enabled', 'true', 'en')
        else:
            cf = Bitstream(content_file, path, bundle=bundle)

        if description != '':
            cf.add_description(description)
        if permissions is not None:
            for p in permissions:
                cf.add_permission(p[0], p[1])
        self.contents.append(cf)
        active_bundle.add_bitstream(cf)

    def enable_entity(self, entity_type: str):
        """
        Enables an item to be a dspace-entity by providing an entity-type.

        :param entity_type: The type of the entity.
        """
        self.add_metadata('dspace.entity.type', entity_type)

    def get_owning_collection(self) -> Collection | None:
        """
        Provides the owning collection of the item, if existing.

        :return: The collection object of the owning collection or None.
        """
        return None if self.collections is None else self.collections[0]

    def get_bundles(self) -> list[Bundle]:
        """
        Returns the bundles used by this item.
        """
        bundles = []
        for c in self.contents:
            if c.bundle not in bundles:
                bundles.append(c.bundle)
        return bundles

    def get_bundle(self, bundle_name: str) -> Bundle | None:
        """
        Returns a specific bundle based on the bundle name or None if the bundle does not exist.

        :param bundle_name: The name of the bundle.
        :return: The bundle object associated with the bundle name, or None if a bundle with this name does not exist.
        """
        for b in self.bundles:
            if b.name == bundle_name:
                return b
        return None

    def get_dspace_object_type(self) -> str:
        return 'Item'

    def __str__(self):
        """
        Creates a string representation of the item object.
        """
        obj_str = super().__str__()
        obj_str = obj_str.replace('DSpace object', 'DSpace item')
        if len(self.relations) > 0:
            obj_str += '\n\tRelations:'
            for r in self.relations:
                obj_str += f'\n\t\t{r}'
        if len(self.contents) > 0:
            obj_str += '\n\tBitstreams:'
            for c in self.contents:
                obj_str += f'\n\t\t{c}'
        if len(self.collections) > 0:
            obj_str += '\n\tCollections:'
            for c in self.collections:
                obj_str += f'\n\t\t{c.uuid if c.uuid != "" else c.handle}'
        return obj_str

    def get_related(self) -> list[DSpaceObject]:
        """
        If this Item is an entity. This method will return a list of related items. If not the list will be empty.

        :return: A list of DSpaceObjects (Items)
        """
        if not self.is_entity():
            return []
        return [r.items[1] for r in self.relations]



import unittest
import dspyce as ds
from dspyce.DSpaceObject import DSpaceObject
from dspyce.Item import Item
from dspyce.Collection import Collection
from dspyce.Community import Community
from dspyce.metadata import MetaData


class DSpaceObjectTest(unittest.TestCase):
    """
    Tests for dspyce.DSpaceObject objects.
    """
    obj: DSpaceObject = DSpaceObject('123445-123jljl1-234kjj', 'doc/12345', 'test-name')

    def test_init_object(self):
        """
        Tests object initialization from DSpaceObject class.
        """
        self.assertEqual(self.obj, DSpaceObject('123445-123jljl1-234kjj'))
        self.assertEqual(self.obj.uuid, '123445-123jljl1-234kjj')
        self.assertEqual(self.obj.handle, 'doc/12345')
        self.assertEqual(self.obj.name, 'test-name')
        self.assertNotEqual(self.obj, DSpaceObject('dddds2-123jljl1-234kjj'))
        self.assertIsInstance(self.obj, DSpaceObject)

    def test_add_metadata(self):
        """
        Tests add_metadata method from DSpaceObject
        """
        self.obj.add_metadata('dc.title', 'hello', 'en')
        self.assertEqual(self.obj.metadata['dc.title'][0].value,
                         'hello')
        self.assertEqual(self.obj.metadata['dc.title'][0].language, 'en')
        self.obj.add_metadata('relation.isAuthorOfPublication.latestForDiscovery',
                              'kidll88-uuid999-duwkke1222', 'en')
        self.assertEqual(self.obj.get_metadata_values('relation.isAuthorOfPublication.latestForDiscovery')[0],
                         'kidll88-uuid999-duwkke1222')
        self.obj.metadata = MetaData({})

    def test_remove_metadata(self):
        self.obj.add_metadata('dc.title', 'hello', 'en')
        self.assertEqual(['hello'], self.obj.get_metadata_values('dc.title'))
        self.obj.remove_metadata('dc.title')
        self.assertIsNone(self.obj.get_metadata_values('dc.title'))
        self.obj.add_metadata('dc.title', 'hello', 'en')
        self.obj.add_metadata('dc.title', 'hallo', 'de')
        self.assertEqual(['hello', 'hallo'], self.obj.get_metadata_values('dc.title'))
        self.obj.remove_metadata('dc.title', 'hallo')
        self.assertEqual(['hello'], self.obj.get_metadata_values('dc.title'))
        self.obj.remove_metadata('dc.title')

    def test_replace_metadata(self):
        self.obj.add_metadata('dc.title', 'hello', 'en')
        self.assertEqual(['hello'], self.obj.get_metadata_values('dc.title'))
        self.obj.replace_metadata('dc.title', 'salut')
        self.assertEqual(['salut'], self.obj.get_metadata_values('dc.title'))
        self.obj.remove_metadata('dc.title')

    def test_object_type(self):
        """
        Test object_type method from DSpaceObject, Item and Community classes.
        """
        self.assertIsNone(self.obj.get_dspace_object_type())
        self.assertEqual(Item().get_dspace_object_type(), 'Item')
        self.assertEqual(Collection().get_dspace_object_type(), 'Collection')
        self.assertEqual(Community().get_dspace_object_type(), 'Community')

    def test_get_identifier(self):
        """
        Tests get_identifier method from DSpaceObject
        """
        self.assertEqual(self.obj.get_identifier(), '123445-123jljl1-234kjj')
        self.assertEqual(Item(handle='doc/12').get_identifier(), 'doc/12')
        self.assertIsNone(Collection().get_identifier())

    def test_equality(self):
        """
        Tests __eq__ method from DSpaceObject
        """
        self.assertTrue(self.obj == DSpaceObject('123445-123jljl1-234kjj'))
        self.assertFalse(self.obj == DSpaceObject('12dsf3445-234kjj'))

    def test_to_dict(self):
        """
        Tests to_dict method from DSpaceObject.
        """
        self.assertEqual({'uuid': '123445-123jljl1-234kjj', 'handle': 'doc/12345', 'name': 'test-name',
                          'metadata': {}}, self.obj.to_dict())

    def test_from_dict(self):
        """
        Tests from_dict function from dspyce.__init__()
        """
        dict_obj = {'uuid': '123445-123jljl1-234kjj', 'handle': 'doc/12345', 'name': 'test-name',
                    'dc.title': [{'value': 'test-title', 'language': 'en'}, {'value': 'Test-Titel', 'language': 'de'}]}
        obj = ds.from_dict(dict_obj)
        self.assertEqual('123445-123jljl1-234kjj', obj.uuid)
        self.assertEqual('doc/12345', obj.handle)
        self.assertEqual('test-name', obj.name)
        self.assertTrue(obj.get_metadata_values('dc.title'), ['test-title', 'Test-Titel'])
        self.assertEqual('en', obj.metadata['dc.title'][0].language)
        self.assertEqual('de', obj.metadata['dc.title'][1].language)
        self.assertRaises(TypeError, ds.from_dict, dict_obj, 'test')

    def test_statistics(self):
        """
        Test add_statistic_report and has_statistics methods from DSpaceObject
        """
        self.assertFalse(self.obj.has_statistics())
        self.obj.add_statistic_report({'TotalViews': 0})
        self.obj.add_statistic_report(None)
        self.obj.add_statistic_report({"TotalDownloads": {'uuid': 'lkjlkjl', "views": 12}})
        self.obj.add_statistic_report({"TotalDownloads": {'uuid': '12345', "views": 12}})
        self.obj.add_statistic_report({"TotalDownloads": {'uuid': 'sd3x33', "views": 15}})
        self.assertEqual(2, len(self.obj.statistic_reports))
        self.assertTrue(isinstance(self.obj.statistic_reports['TotalDownloads'], list))
        self.assertTrue(self.obj.has_statistics())


if __name__ == '__main__':
    unittest.main()



import unittest
import dspyce.metadata as md


class MetadataTest(unittest.TestCase):
    """
    Tests for metadata package.
    """
    mdValue = md.MetaDataValue('Hello World', 'en')
    mdList = md.MetaData({'dc.title': [mdValue]})

    def test_equal(self):
        """
        Test __eq__ method from MetaData class.
        """
        self.assertTrue(self.mdValue == md.MetaDataValue('Hello World', 'en'))
        self.assertFalse(self.mdValue == md.MetaDataValue('Smith, Adam', 'en'))
        self.assertRaises(TypeError, self.mdValue.__eq__, 'abc')

    def test_to_string(self):
        """
        Tests __str__ method from MetaData class.
        """
        self.assertEqual('en:\tHello World', str(self.mdValue))
        self.assertEqual('dc.title:\n\ten:\tHello World', str(self.mdList))

    def test_add_value(self):
        """
        Tests add_value method from MetaData class.
        """
        self.mdValue.set_value('Smith, Adam')
        self.assertEqual('Smith, Adam', self.mdValue.value)
        self.mdValue.set_value('Hello World')
        self.mdList['dc.contributor.author'] = md.MetaDataValue('Smith, Adam')
        self.assertEqual('Smith, Adam', self.mdList['dc.contributor.author'][0].value)
        self.mdList.pop('dc.contributor.author')
        self.assertIsNone(self.mdList.get('dc.contributor.author'))
        self.assertRaises(TypeError, self.mdList.__setitem__, 'dc.title', 'xyz')
        self.assertRaises(KeyError, self.mdList.__setitem__, 'hello', md.MetaDataValue('test'))

    def test_to_dict(self):
        """
        Test the dict() implementation of the MetaDataValue class.
        """
        self.assertEqual({'value': 'Hello World', 'language': 'en'}, dict(self.mdValue))

    def test_pop_value(self):
        obj = md.MetaData({'dc.title': [md.MetaDataValue('Hello World', 'en')]})
        self.assertEqual(md.MetaDataValue('Hello World', 'en'), obj['dc.title'][0])
        obj.pop('dc.title')
        self.assertIsNone(obj.get('dc.title'))


if __name__ == '__main__':
    unittest.main()



from .DSpaceObject import DSpaceObject
from .Community import Community


class Collection(DSpaceObject):
    """
    The class Collection represents a DSpace collection, containing different Items and having a parent community.
    """
    community: Community

    def __init__(self, uuid: str = '', handle: str = '', name: str = '', community: Community = None, ):
        super().__init__(uuid, handle, name)
        self.community = community

    def get_parent_community(self) -> Community:
        """
        Returns the parent community of the Collection, if existing.

        :return: Community object or None, if none existing.
        """
        return self.community

    def get_dspace_object_type(self) -> str:
        return 'Collection'



import unittest

import dspyce as ds


class RelationTest(unittest.TestCase):
    """
    Tests methods from Relation class.
    """
    relation = ds.Relation('isAuthorOfPublication',
                           (ds.Item('abc'),
                            ds.Item('dfg')), 2)

    def test_str(self):
        """
        Tests __str__ method.
        """
        self.assertEqual('abc:relation.isAuthorOfPublication:dfg', str(self.relation))

    def test_eq(self):
        """
        Tests __eq__ method.
        """
        self.assertEqual(ds.Relation('isAuthorOfPublication', (ds.Item(), ds.Item()),
                                     2), self.relation)
        self.assertNotEqual(ds.Relation('isAuthorOfPublication', (ds.Item(), ds.Item())),
                            self.relation)

    def test_set_relation_type(self):
        """
        Tests set_relation_type method.
        """
        self.relation.set_relation_type(1)
        self.assertEqual(1, self.relation.relation_type)


if __name__ == '__main__':
    unittest.main()



# entities/__init__.py
"""
Python package for analysing and working with entity modells in DSpace.
This package is mostly used for drawing entity graphs based on the relationship-types.xml or a REST endpoint.

To draw an entity modell based on a RestApi, you must run the following code:
>>> from dspyce import entities
>>> entities.from_rest_api('https://sandbox.dspace.org/server/api').draw_graph()
"""
from .models import EntityModell
from .models import from_relationship_file, from_rest_api
from .models import check_entities_rest



import re
import logging

import dspyce.entities
import dspyce.saf
import dspyce.rest
import dspyce.statistics
from dspyce.DSpaceObject import DSpaceObject
from dspyce.Item import Item
from dspyce.Collection import Collection
from dspyce.Community import Community
from dspyce.Relation import Relation


def from_dict(obj_dict: dict, obj_type: str = None) -> DSpaceObject | Item | Community | Collection:
    """
    Creates a new DSpaceObject from a given dictionary.
    Example for dict:
        {
          'uuid': <uuid>,

          'handle': <handle>,

          'name': <name>,

          '<dc.[...]>': <value>,

          '<[metadata-tag]>': <value> | list[<value>] | dict[lang: list[<value>]], # A list of metadata tags.

          'relation.<relation-name>': <uuid>, # A number of relationships with the connected uuids for items.

          'parent_community': <uuid>, # for possible parent communities if the obj_type is collection or community

          'collection': <uuid>, # A owning collection if the object is an Item.
        }

    >>> from_dict({'uuid': 'lkj-123-123-jlkjld', 'name': 'example',}, 'item')
    DSpaceObject('lkj-123-123-jlkjld', name='item')

    :param obj_dict: The dictionary containing the object information.
    :param obj_type: The type of DspaceObject. Must be one of (community, item, collection).
    :return: A DSpaceObject.
    """
    obj = None
    match obj_type:
        case None:
            obj = DSpaceObject(obj_dict['uuid'] if 'uuid' in obj_dict.keys() else '',
                               obj_dict['handle'] if 'handle' in obj_dict.keys() else '',
                               obj_dict['name'] if 'name' in obj_dict.keys() else '')
        case 'item':
            obj = Item(obj_dict['uuid'] if 'uuid' in obj_dict.keys() else '',
                       obj_dict['handle'] if 'handle' in obj_dict.keys() else '',
                       obj_dict['name'] if 'name' in obj_dict.keys() else '')
            if 'collection' in obj_dict.keys():
                obj.collections = [Collection(obj_dict['collection'])]
            relationships = filter(lambda x: re.search(r'^relation.', x), obj_dict.keys())
            for r in relationships:
                obj.add_relation(r.split('.')[1], obj_dict[r])

        case 'community':
            obj = Community(obj_dict['uuid'] if 'uuid' in obj_dict.keys() else '',
                            obj_dict['handle'] if 'handle' in obj_dict.keys() else '',
                            obj_dict['name'] if 'name' in obj_dict.keys() else '')
            if 'parent_community' in obj_dict.keys():
                obj.parent_community = Community(obj_dict['parent_community'])
        case 'collection':
            obj = Collection(obj_dict['uuid'] if 'uuid' in obj_dict.keys() else '',
                             obj_dict['handle'] if 'handle' in obj_dict.keys() else '',
                             obj_dict['name'] if 'name' in obj_dict.keys() else '')
            if 'parent_community' in obj_dict.keys():
                obj.parent_community = Community(obj_dict['parent_community'])
        case _:
            raise TypeError('The obj_type parameter must be one of (item, collection, community or None)'
                            f'but got {obj_type}')
    for key in filter(lambda x: re.search(r'[a-zA-Z0-9\-]\.[a-zA-Z0-9\-](\.[a-zA-Z0-9\-])?', x),
                      obj_dict.keys()):
        if not isinstance(obj_dict[key], list) and not isinstance(obj_dict[key], dict):
            obj.add_metadata(tag=key, value=obj_dict[key])
        elif isinstance(obj_dict[key], list):
            for o in obj_dict[key]:
                if isinstance(o, dict) and 'value' in o.keys():
                    obj.add_metadata(tag=key, value=o['value'],
                                     language=o['language'] if 'language' in o.keys() else None)
                else:
                    obj.add_metadata(tag=key, value=o)
        else:
            val = obj_dict[key]
            obj.add_metadata(tag=key, value=val['value'], language=val['language'] if 'language' in val.keys() else None)
    return obj



# entities/models.py
"""
Python package for analysing and working with entity modells in DSpace.
This module contains the EntityModell class and three functions to work with entities. You can create a new EntityModell
object based on the relationship-types.xml file or a given REST-endpoint and you can check whether a given endpoint
has an entity modell enabled or not.
"""
from bs4 import BeautifulSoup
import matplotlib.pyplot as plt
import networkx as nx
from dspyce.rest import RestAPI


class EntityModell:
    """
    The EntityModell class provides methods and attributes to work with entity modells. An object of this class
    represents entities and relationships of a DSpace repository.
    """

    entities: list[str]
    """A list of existing entities."""
    relations: list[str]
    """A list of the names of existing relations."""
    entity_model: nx.Graph
    """The networkx graph representing the entity model."""

    def __init__(self):
        """
        Creates a new object of the entity model.
        """
        self.entities = []
        self.relations = []
        self.entity_modell = nx.MultiDiGraph()

    def has_entity(self, entity: str) -> bool:
        """
        Checks if the current entity already exits.

        :param entity: The name of the entity to check for.
        :return: True if the entity exists already in the entity model.
        """
        return entity in self.entities

    def has_relation(self, relation: str) -> bool:
        """
        Checks if the current relation already exits.

        :param relation: The name of the relation to check for.
        :return: True if the relation exists already in the entity model.
        """
        return relation in self.relations

    def add_entity(self, entity: str):
        """
        Adds a new entity to the entity modell.

        :param entity: The name of the entity to add to the modell.
        :raises AttributeError: If the entity already exists.
        """
        if not self.has_entity(entity):
            self.entities.append(entity)
            self.entity_modell.add_node(entity)
        else:
            raise ValueError(f"The entity {entity} already exists.")

    def add_relation(self, from_entity: str, to_entity: str, relation_name: str):
        """
        Adds a relation between two entities. Relations are always directional.

        :param from_entity: The name of the starting entity.
        :param to_entity: The name of the target entity.
        :param relation_name: The name of the relation between the two entities.
        :raises AttributeError: If the relation already exists.
        """
        if not self.has_relation(relation_name):
            self.entity_modell.add_edge(from_entity,
                                        to_entity,
                                        label=relation_name)
        else:
            raise ValueError(f"The relation {relation_name} already exists.")

    def draw_graph(self, show: bool = True, path: str = None):
        """
        Draws the current entity model as a graph using matplotlib.

        :param show: If true the graph will be displayed in a plt figure.
        :param path: The path and file_name where the graph should be saved. If None, the graph won't be saved.
        """
        plt.axis('off')
        fig, ax = plt.subplots(figsize=(2.13*len(self.entities)+2,
                                        1.6*len(self.entities)+2))
        ax.axis('off')
        pos = nx.spring_layout(self.entity_modell, scale=2)
        # con_style = 'arc3, rad=0.1'
        con_style = [f"arc3,rad={r}" for r in [0.2, 0.6, 0.9, 1.2]]
        nx.draw_networkx(self.entity_modell, pos,
                         node_size=1500, alpha=0.9, node_shape='s',
                         edge_color='grey', connectionstyle=con_style,
                         min_source_margin=30, min_target_margin=30, label='labels',
                         verticalalignment='center_baseline',
                         font_color='black', bbox={"alpha": 0.7, "color": "white"}, ax=ax
                         )
        nx.draw_networkx_edge_labels(
            self.entity_modell, pos, nx.get_edge_attributes(self.entity_modell, 'label'),
            label_pos=0.5, font_color='black', bbox={"alpha": 0.9, "color": "white"},
            connectionstyle=con_style, ax=ax)
        if path is not None:
            fig.savefig(path)
        if show:
            fig.show()

    def get_relation_list(self) -> list[tuple[tuple[str, str], str]]:
        """
        Returns a list of all relations and the connected entity types.

        :return: Returns the list in the following format [((left_type, right_type), relation_name), ...]
        """
        return [((l, r), attr['label']) for l, r, attr in self.entity_modell.edges(data=True)]


def from_relationship_file(path: str) -> EntityModell:
    """
    Loads a EntityModell from a given DSpace relationship file, aka `relationship-types.xml`.

    :param path: The path to the relationship xml file.
    :return: A complete EntityModell object based on the content provided from the relationship file.
    """
    with open(path, 'r', encoding='utf8') as f:
        bs = BeautifulSoup(f.read(), 'xml')
    relationships = bs.relationships
    md = EntityModell()
    for r in relationships.contents:
        if r.name == 'type':
            left_type = r.leftType.string
            right_type = r.rightType.string
            if not md.has_entity(left_type):
                md.add_entity(left_type)
            if not md.has_entity(right_type):
                md.add_entity(right_type)
            md.add_relation(right_type, left_type, r.leftwardType.string)
            md.add_relation(left_type, right_type, r.rightwardType.string)
    return md


def from_rest_api(url: str) -> EntityModell:
    """
    Creates an EntityModell object based on the information provided from the REST API

    :param url: The url of the used RestApi.
    :return: A complete EntityModell object based on the RestAPI provided.
    """
    rest = RestAPI(url)
    entity_objects = rest.get_paginated_objects('core/entitytypes', 'entitytypes')
    entity_objects = list(filter(lambda x: x['label'] != 'none', entity_objects))
    if len(entity_objects) == 0:
        raise ValueError(f'No entity types found in instance "{url}"')
    em = EntityModell()
    [em.add_entity(e['label']) for e in entity_objects]
    relations = rest.get_paginated_objects(f'core/relationshiptypes', 'relationshiptypes')
    for r in relations:
        leftward_type = r['leftwardType']
        rightward_type = r['rightwardType']
        left_type = rest.get_api(f'core/entitytypes/{r["_links"]["leftType"]["href"].split("/")[-1]}')['label']
        right_type = rest.get_api(f'core/entitytypes/{r["_links"]["rightType"]["href"].split("/")[-1]}')['label']
        if not em.has_relation(leftward_type):
            em.add_relation(right_type, left_type, leftward_type)
        if not em.has_relation(rightward_type):
            em.add_relation(left_type, right_type, rightward_type)
    return em


def check_entities_rest(url: str) -> bool:
    """
    Checks if the given rest endpoint has entities enabled.

    :param url: The url of the rest-API
    """
    entity_objects = RestAPI(url).get_paginated_objects('core/entitytypes', 'entitytypes')
    return len(list(filter(lambda x: x['label'] != 'none', entity_objects))) > 0



from ..rest import RestAPI
from ..DSpaceObject import DSpaceObject

REPORT_TYPES: tuple = ('TotalVisits', 'TotalVisitsPerMonth', 'TotalDownloads', 'TopCountries', 'TopCities')
"""The allowed report_type values."""


def get_point_views(point: dict) -> int | None:
    """
    Return the views value from a point dict, retrieved by the RestAPI of DSpace instance.

    :param point: The dictionary contain the points' data.
    :return: The views value, if existing, else None.
    :raises TypeError: If the parameter point is not form type dict.
    """
    if not isinstance(point, dict):
        raise TypeError(f'The parameter point must be a dictionary! But found: {type(point)}')
    if isinstance(point.get('values'), dict) and 'views' in point['values'].keys():
        return point['values']['views']

    return None


def download_statistics_to_object(object_uuid: str, report_type: str, rest_api: RestAPI) -> dict | None:
    """
    Retrieves a statistics report from a given RestAPI endpoint connected to the given object_id.

    :param object_uuid: The uuid of the object to get the report from.
    :param report_type: The report type to retrieve. (Must be one of REPORT_TYPES)
    :param rest_api: A given rest API to retrieve the data from.
    :return: The report as a dict object containing the information.
    :raise ValueError: If the report type is incorrect.
    """
    if report_type not in REPORT_TYPES:
        raise ValueError(f'The report_type must be one of {REPORT_TYPES}, but got "{report_type}"')
    request_id = f'{object_uuid}_{report_type}'
    json_resp = rest_api.get_api(f'statistics/usagereports/{request_id}')
    if json_resp is None:
        return None
    if 'points' not in json_resp.keys() or len(json_resp['points']) <= 0:
        return None
    match report_type:
        case 'TotalVisits':
            return {report_type: get_point_views(json_resp['points'][0])}
        case 'TotalDownloads':
            downloads = []
            for p in json_resp['points']:
                downloads.append({'uuid': p['id'], 'label': p['label'], 'downloads': get_point_views(p)})
            return {report_type: downloads}
        case 'TotalVisitsPerMonth':
            return {report_type: {p['label']: get_point_views(p) for p in json_resp['points']}}
    return None


def download_statistics(object_list: list[str] | list[DSpaceObject], report_type: str,
                        rest_api: RestAPI) -> list[dict] | list[DSpaceObject]:
    """
    Retrieves statistic reports for list of DSpaceObject base on their uuids. Adds the reports directly to the objects
    if a list of DSpaceObjects is provided.

    :param object_list: A list of object uuids or DSpaceObject objects to retrieve the data for.
    :param report_type: The type of report to retrieve.
    :param rest_api: The rest_api to retrieve the statistics from.
    :return: A list of Report objects, if object_list is a list of uuids, else: returns a list of DSpaceObjects with
    added statistics.
    """
    if len(object_list) > 0 and isinstance(object_list[0], str):
        return [download_statistics_to_object(str(o), report_type, rest_api) for o in object_list]

    for o in object_list:
        o: DSpaceObject
        tmp = download_statistics_to_object(o.uuid, report_type, rest_api)
        o.add_statistic_report(tmp)
        del tmp
    return object_list


def all_statistics_to_object(obj: str | DSpaceObject, rest_api: RestAPI) -> list[dict] | DSpaceObject:
    """
    Returns all currently available reports for an object as a list or the updated DSpaceObject object if the object
    parameter is of type DSpaceObject.

    :param obj: The uuid of the object to get the report from or the corresponding DSpaceObject.
    :param rest_api: A given rest API to retrieve the data from.
    :return: A Report object containing the information.
    :raise TypeError: If type(object) is not DSpaceObject
    """
    if isinstance(obj, DSpaceObject):
        stats = list(filter(lambda x: x is not None,
                            [download_statistics_to_object(obj.uuid,
                                                           report_type, rest_api) for report_type in REPORT_TYPES]))
        obj.add_statistic_report(stats)
        del stats
        return obj
    if isinstance(obj, str):
        return list(filter(lambda x: x is not None,
                           [download_statistics_to_object(obj,
                                                          report_type, rest_api) for report_type in REPORT_TYPES]))

    raise TypeError(f"The obj type must be either DSpaceObject or str, but got {type(obj)}!")


def all_statistics(object_list: list[str] | list[DSpaceObject],
                   rest_api: RestAPI) -> list[dict] | list[DSpaceObject]:
    """
    Retrieves all statistic reports for a given list of DSpaceObject based on their uuids. Adds the reports directly to
    the objects if a list of DSpaceObjects is provided.

    :param object_list: A list of object uuids or DSpaceObject objects to retrieve the data for.
    :param rest_api: The rest_api to retrieve the statistics from.
    :return: A list of Report objects, if object_list is a list of uuids, else: returns a list of DSpaceObjects with
    added statistics.
    """
    if len(object_list) > 0 and isinstance(object_list[0], str):
        reports = []
        for o in object_list:
            o: str
            reports += all_statistics_to_object(o, rest_api)
        return reports

    for o in object_list:
        o: DSpaceObject
        o.add_statistic_report(all_statistics_to_object(o.uuid, rest_api))
    return object_list



import logging
import unittest
from dspyce import saf
from dspyce import Item, Collection
from dspyce.metadata import MetaDataValue
import os


class SAFTest(unittest.TestCase):
    def test_saf(self):
        saf.saf_write.LOG_LEVEL = logging.DEBUG
        created_dir = False
        if 'test_data' not in os.listdir():
            os.mkdir('test_data')
            created_dir = True
        item = Item(handle='123456789/3', collections=[Collection(handle='123456789/1'),
                                                       Collection(handle='123456789/2')])
        item.add_metadata('dc.title', 'Test Publication', 'en')
        item.add_metadata('dc.type', 'article', 'en')
        item.add_metadata('dc.contributor.author', 'Smith, Adam')
        item.enable_entity('Publication')
        item.add_metadata('local.test.field', 'Nothing', 'en')
        saf.create_saf_package(item, 0, './test_data/')
        self.assertIn('archive_directory', os.listdir('./test_data/'))
        self.assertIn('item_0', os.listdir('./test_data/archive_directory'))
        self.assertEqual(
            ['collections', 'dublin_core.xml', 'handle', 'metadata_dspace.xml', 'metadata_local.xml'].sort(),
            os.listdir('./test_data/archive_directory/item_0').sort())
        self.assertRaises(FileExistsError,  saf.create_saf_package, item, 0, './test_data/')

        saf_item = saf.read_saf_packages('test_data/archive_directory/')[0]
        self.assertIn(MetaDataValue('Test Publication', 'en'), saf_item.metadata['dc.title'])
        self.assertIn(MetaDataValue('Nothing', 'en'), saf_item.metadata['local.test.field'])
        self.assertIn(MetaDataValue('Publication'), saf_item.metadata['dspace.entity.type'])
        self.assertTrue(saf_item.is_entity())
        self.assertIn(Collection(handle='123456789/2'), saf_item.collections)
        self.assertEqual(Collection(handle='123456789/1'), saf_item.get_owning_collection())
        self.assertEqual('123456789/3', saf_item.handle)

        for file in os.listdir('./test_data/archive_directory/item_0'):
            logging.debug(f'Remove file: ./test_data/archive_directory/item_0/{file}')
            os.remove(f'./test_data/archive_directory/item_0/{file}')
        os.rmdir(f'./test_data/archive_directory/item_0')
        logging.debug(f'Remove directory: ./test_data/archive_directory/item_0')
        os.rmdir(f'./test_data/archive_directory')
        logging.debug(f'Remove directory: ./test_data/archive_directory')
        if created_dir:
            os.rmdir('./test_data')
            logging.debug(f'Remove directory: ./test_data')



# saf/saf_read.py
"""
Python module for reading out saf_packages to create dspyce.Item objects

1. read_saf_package() -> Reads a single saf package and creates the Item object.
2. read_saf_packages() -> Reads all found saf packages in the provided path and returns the Item objects based on
the read_saf_packages() function.
"""

import os
import re
import uuid
from bs4 import BeautifulSoup

from dspyce.Item import Item
from dspyce.Collection import Collection
from dspyce.bitstreams import Bundle


def read_saf_package(path: str) -> Item:
    """
    Reads a single saf package and creates the Item based on the found metadata, bitstreams, collections, handles and
    relationships.

    :param path: The path where to find the directory of the saf-package.
    :return: The Item object for the saf-package.
    """
    path = path[:-1] if path.endswith('/') else path
    files = os.listdir(path)
    metadata_files = filter(lambda x: re.search(r'((metadata_[a-zA-Z\-]+)|(dublin_core))\.xml$', x), files)
    further_information: dict[str: list] = {'contents': [], 'collections': [], 'relationships': []}
    handle = ''
    for file_name in ('contents', 'collections', 'relationships'):
        if file_name in files:
            with open(f'{path}/{file_name}', encoding='utf-8') as f:
                further_information[file_name] = f.read().splitlines()
    if 'handle' in files:
        with open(f'{path}/handle', encoding='utf-8') as f:
            handle = f.read().splitlines()[0].strip()
    item = Item(handle=handle)
    for md_file in metadata_files:
        with open(f'{path}/{md_file}', encoding='utf-8') as f:
            xml_content = f.read()
        bs = BeautifulSoup(xml_content, 'xml').find('dublin_core')
        schema = bs.get('schema')
        if schema is None:
            schema = 'dc'
        for field in bs.find_all('dcvalue'):
            element = field.get('element')
            qualifier = field.get('qualifier')
            qualifier = None if qualifier is None or qualifier == 'none' else qualifier
            tag = f'{schema}.{element}' + (f'.{qualifier}' if qualifier is not None else '')
            lang = field.get('language')
            item.add_metadata(tag, field.get_text(), lang)

    for c in further_information['collections']:
        try:
            item.add_collection(Collection(uuid=str(uuid.UUID(c))))
        except ValueError:
            item.add_collection(Collection(handle=c))

    for r in filter(lambda x: x.strip() != '', further_information['relationships']):
        relation = r.split(' ')
        item.add_relation(relation[0].replace('relation.', ''), relation[1])
    for b in filter(lambda x: x.strip() != '', further_information['contents']):
        bitstream = b.split('\t')
        name = bitstream[0]
        attributes = {i.split(':')[0]: i.split(':')[1] for i in bitstream[1:]}
        description = attributes.get('description') if attributes.get('description') is not None else ''
        bundle = attributes.get('bundle') if attributes.get('bundle') is not None else Bundle.DEFAULT_BUNDLE
        permissions = [(p.split(' ')[0], p.split(' ')[1])
                       for p in attributes.get('permissions')] if attributes.get('permissions') is not None else None

        iiif = 'iiif-label' in attributes.keys()
        item.add_content(name, f'{path}', description, bundle=bundle, permissions=permissions, iiif=iiif,
                         iiif_toc=attributes['iiif-toc'] if 'iiif-toc' in attributes.keys() else '')
    return item


def read_saf_packages(path: str) -> list[Item]:
    """
    Read all SAF-packages stored in the directory provided by path.

    :param path: The directory where the saf_packages are stored in.
    :return: A list of Item objects built based on the information provided by the SAF-packages.
    """
    return [read_saf_package(f'{path}/{directory}') for directory in os.listdir(path)]



# saf/__init__.py
"""
Python package for creating and reading saf packages for DSpace Item-imports or -exports.
"""

from .saf_write import saf_packages, create_saf_package
from .saf_read import read_saf_packages, read_saf_package



import json
import logging
import requests
from requests.exceptions import InvalidJSONError

from ..DSpaceObject import DSpaceObject
from ..Item import Item
from ..Community import Community
from ..Collection import Collection
from ..Relation import Relation
from ..bitstreams import Bundle, Bitstream, IIIFBitstream
from ..metadata import MetaData


def json_to_object(json_content: dict) -> DSpaceObject | Item | Community | Collection | None:
    """
    Converts a dict based on REST-format in to a DSpace Object.

    :param json_content: The json content in a dict format.
    :return: A DSpaceObject object.
    """
    uuid = json_content['uuid']
    name = json_content['name']
    handle = json_content['handle']
    metadata = json_content['metadata']
    doc_type = json_content['type']
    if json_content is None:
        return None
    match doc_type:
        case 'community':
            obj = Community(uuid, handle=handle, name=name)
        case 'collection':
            obj = Collection(uuid, handle=handle, name=name)
        case 'item':
            obj = Item(uuid, handle=handle, name=name)
        case _:
            obj = DSpaceObject(uuid, handle, name)
    for m in metadata.keys():
        for v in metadata[m]:
            value = v['value']
            lang = v['language'] if 'language' in v.keys() else None
            obj.add_metadata(tag=m, value=value, language=lang)
    return obj


def object_to_json(obj: DSpaceObject) -> dict:
    """
    Converts a DSpaceObject class into dict based on a DSpace-Rest format
    :param obj: The object to convert.
    :return: A dictionary in the REST-format.
    """
    uuid = obj.uuid
    handle = obj.handle
    name = obj.name
    obj_type = obj.get_dspace_object_type().lower()
    metadata = {key: [dict(v) for v in obj.metadata[key]] for key in obj.metadata.keys()}

    json_object = {}
    if uuid is not None and uuid != '':
        json_object['uuid'] = uuid
    if handle is not None and handle != '':
        json_object['handle'] = handle
    if name is not None and name != '':
        json_object['name'] = name
    if obj_type is not None and obj_type != '':
        json_object['type'] = obj_type
    if isinstance(obj, Item):
        obj: Item
        json_object['inArchive'] = obj.in_archive
        json_object['discoverable'] = obj.discoverable
        json_object['withdrawn'] = obj.withdrawn
        if obj.is_entity():
            json_object['entityType'] = obj.get_entity_type()
    if obj.handle is not None and obj.handle != '':
        json_object['handle'] = obj.handle
    json_object['metadata'] = metadata
    return json_object


class RestAPI:
    """
    The class RestAPI represents the REST API of a DSpace 7 backend. It helps to get, push, update or remove Objects,
    Bitstreams, Relations, MetaData and other from a DSpace Instance.
    """
    api_endpoint: str
    """The address of the api_endpoint."""
    username: str
    """The username of the user communicating to the endpoint."""
    password: str
    """The password of the user communicating with the endpoint."""
    session: requests.sessions.Session
    """The active session."""
    authenticated: bool = False
    """Provides information about the authentication status."""
    dspace_version: str
    """The dspace version used by the API endpoint."""

    def __init__(self, api_endpoint: str, username: str = None, password: str = None,
                 log_level: int | str = logging.INFO, log_file: str = None):
        """
        Creates a new object of the RestAPI class using

        :param api_endpoint: The api endpoint to connect to. (For example https://demo.dspace.org/server/api)
        :param username: The username of the user for authentication.
        :param password: The password of the user for authentication.
        :param log_level: The log_level used for Logging. Must be string or integer. The strings must be one of
            the following: DEBUG, INFO, WARNING, ERROR, CRITICAL. Default is INFO.
        :param log_file: A possible name and path of the log file. If None provided, all output will be logged to the
            console.
        """
        log_level_types = {'DEBUG': logging.DEBUG, 'INFO': logging.INFO, 'WARNING': logging.WARNING,
                           'ERROR': logging.ERROR, 'CRITICAL': logging.CRITICAL}
        if isinstance(log_level, str) and log_level.upper() not in log_level_types.keys():
            raise TypeError(f"Invalid log level: {log_level}. Must be one of DEBUG, INFO, WARNING, ERROR, CRITICAL.")
        elif isinstance(log_level, str):
            log_level = log_level_types[log_level]
        logging.basicConfig(level=log_level, filename=log_file, encoding='utf8',
                            format='%(asctime)s - %(levelname)s: %(message)s')
        self.session = requests.Session()
        self.api_endpoint = api_endpoint
        endpoint_info = RestAPI.get_endpoint_info(api_endpoint)
        if endpoint_info is None:
            logging.critical(f'Couldn\'t reach the api_endpoint with the address "{api_endpoint}".')
            raise requests.exceptions.ConnectionError(f'Could not reach the endpoint {api_endpoint}.')
        else:
            self.dspace_version = endpoint_info['dspaceVersion']
        self.username = username
        self.password = password
        self.req_headers = {'Content-type': 'application/json', 'User-Agent': 'Python REST Client'}
        if username is not None and password is not None:
            self.authenticated = self.authenticate_api()

    @staticmethod
    def get_endpoint_info(api: str) -> dict[str, str] | None:
        """
        Checks if the request endpoint is reachable and returns name, ui_address, server_address and dspace-Version

        :param api: The url of the API endpoint.
        """
        logging.debug(f'Checking endpoint status.')
        req = requests.get(api)
        if req.status_code in (204, 201, 200):
            try:
                logging.debug(f'Established connection with the endpoint. Status code: {req.status_code}')
                resp = req.json()
                logging.info(f'Connection with endpoint "{api}" established. Instance-name: "{resp["dspaceName"]}",'
                             f'UI-address: "{resp["dspaceUI"]}", Server-address: "{resp["dspaceServer"]}",'
                             f'DSpace-Version: "{resp["dspaceVersion"]}"')
                return resp
            except json.decoder.JSONDecodeError:
                logging.critical('The reponse could not be parsed. Are you sure, that you\'ve provided the url of a '
                                 'valid dspace 7.x endpoint?')
                return None
        else:
            logging.warning(f'Problems with establishing a connection to the endpoint. Status code: {req.status_code}')
            return None

    def update_csrf_token(self, req: requests.models.Request | requests.models.Response):
        """
        Update the csrf_token based on the current requests.

        :param req: The current request to check the token from.
        """
        if 'DSPACE-XSRF-TOKEN' in req.headers:
            csrf = req.headers['DSPACE-XSRF-TOKEN']
            self.session.headers.update({'X-XSRF-Token': csrf})
            self.session.cookies.update({'X-XSRF-Token': csrf})

    def authenticate_api(self) -> bool:
        """
        Authenticates to the REST-API

        :return: True, if the authentication worked.
        """
        logging.info(f'Trying to authenticate against the REST-API "{self.api_endpoint}", with user {self.username}')
        auth_url = f'{self.api_endpoint}/authn/login'
        req = self.session.post(auth_url)
        self.update_csrf_token(req)
        req = self.session.post(auth_url, data={'user': self.username, 'password': self.password})
        if 'Authorization' in req.headers:
            self.session.headers.update({'Authorization': req.headers.get('Authorization')})
        # Check if authentication was successfully:
        auth_session = self.session.get(auth_url.replace('login', 'status'))
        try:
            auth_status = auth_session.json()
            if 'authenticated' in auth_status and auth_status['authenticated'] is True:
                logging.info(f'The authentication as "{self.username}" was successfully')
                return True
        except requests.exceptions.JSONDecodeError as e:
            logging.error('Problem with authenticating to the api.')
            logging.error(auth_session)
            logging.exception(e)

        logging.critical('The authentication was unsuccessful.')
        return False

    def get_api(self, endpoint: str, params: dict = None) -> dict | None:
        """
        Performs a get request to the api based on a given string endpoint returns the JSON response if successfully.

        :param endpoint: The endpoint information: aka https://self.api_endpoint/<endpoint>
        :param params: A list of additional parameters to pass to the endpoint.
        :return: The json response as a dict.
        """
        endpoint = endpoint if endpoint[0] != '/' else endpoint[1:]
        url = f'{self.api_endpoint}/{endpoint}'
        req = self.session.get(url, params=params if params is not None else {})
        self.update_csrf_token(req)
        if req.status_code in (204, 201, 200):
            logging.debug(f'Successfully performed GET request to endpoint {endpoint}')
            return req.json()
        if req.status_code == 404:
            logging.warning(f'Object behind "{url}" does not exists.')
            logging.warning(req.json())
            return None
        logging.error(f'Problem with performing GET request to endpoint {endpoint}.')
        logging.error(req)

        raise requests.exceptions.RequestException(f'Could not get item with from endpoint: {url}')

    def post_api(self, url: str, json_data: dict, params: dict) -> dict:
        """
        Performs a post action on the RestAPI endpoint.
        """
        req = self.session.post(url)
        self.update_csrf_token(req)
        logging.debug(f'Performing POST request in "{url}" with params({params}):{json_data}')
        try:
            resp = self.session.post(url, json=json_data, headers=self.req_headers, params=params)
        except InvalidJSONError as e:
            logging.error(f'Invalid json format in the query data: {json_data}')
            raise e
        if resp.status_code in (201, 200):
            # Success post request
            json_resp = resp.json()
            logging.info(f'Successfully added object with uuid: {json_resp["uuid"]}')
            return json_resp
        logging.error(f'Could not POST content: {json_data}.\n\tWith params: {params}\n\tOn endpoint: {url}')
        logging.error(f'Statuscode: {resp.status_code}')
        raise requests.exceptions.RequestException(f'\nStatuscode: {resp.status_code}\n'
                                                   f'Could not post content: \n\t{json_data}'
                                                   f'\nWith params: {params}\nOn endpoint:\n\t{url}')

    def patch_api(self, url: str, json_data: list, params: dict = None) -> dict | None:
        """
        Sends a patch request to the api in order to update, add or remove metadata information.

        :param url: The url of the api, where to replace the metadata.
        :param json_data: The data object containing action information.
        :param params: Additional params for the operation.
        :return: The JSON response of the server, if the operation was successfully.
        :raise RequestException: If the JSON response doesn't have the status code 200 or 201
        """
        url = f'{self.api_endpoint}/{url}' if self.api_endpoint not in url else url
        logging.debug(f'Performing PATCH request in "{url}" with params({params}):{json_data}')
        req = self.session.patch(url)
        self.update_csrf_token(req)
        resp = self.session.patch(url, json=json_data, headers=self.req_headers)

        if resp.status_code in (201, 200):
            # Success post request
            json_resp = resp.json()
            logging.info(f'Successfully updated object with uuid: {json_resp["uuid"]}.')
            return json_resp

        if resp.status_code == 204:
            operation = [int((i["op"] if "op" in i.keys() else '') == "remove") - 1 for i in json_data]
            if sum(operation) == 0:
                logging.info('Successfully deleted objects.')
                return None

        logging.error(f'Could not PATCH content: {json_data}.\n\tWith params: {params}\n\tOn endpoint: {url}')
        logging.error(f'Statuscode: {resp.status_code}')
        raise requests.exceptions.RequestException(f'\nStatuscode: {resp.status_code}\n'
                                                   f'Could not put content: \n\t{json_data}'
                                                   f'\nWith params: {params}\nOn endpoint:\n\t{url}')

    def put_api(self, url: str, data: list | dict | str, params: dict = None, content_type: str = None) -> None:
        """
        Sends a PUT request to the api.

        :param url: The path in the api.
        :param data: The data object containing action information.
        :param params: Additional params for the operation.
        :param content_type: The content_type of the data. The default ist self.request.headers['Content-Type']
        :raise RequestException: If the JSON response doesn't have the status code 200 or 201
        """
        url = f'{self.api_endpoint}/{url}' if self.api_endpoint not in url else url
        logging.debug(f'Performing PUT request in "{url}" with params({params}):{data}')
        req = self.session.put(url)
        self.update_csrf_token(req)
        headers = self.req_headers
        if content_type != headers['Content-type']:
            headers['Content-type'] = content_type
        resp = self.session.put(url, data=data, headers=headers)

        if resp.status_code in (204, 200):
            # Success put request
            logging.info(f'Successfully performed put request on endpoint {url}.')
            return

        logging.error(f'Could not PUT content: {data}.\n\tWith params: {params}\n\tOn endpoint: {url}')
        logging.error(f'Statuscode: {resp.status_code}')
        raise requests.exceptions.RequestException(f'\nStatuscode: {resp.status_code}\n'
                                                   f'Could not put content: \n\t{data}'
                                                   f'\nWith params: {params}\nOn endpoint:\n\t{url}')

    def delete_api(self, url: str, params: dict = None, content_type: str = None) -> None:
        """
        Sends a DELETE request to the api.

        :param url: The path in the api.
        :param params: Additional params for the operation.
        :param content_type: The content_type of the data. The default ist self.request.headers['Content-Type']
        :raise RequestException: If the JSON response doesn't have the status code 200 or 201
        """
        url = f'{self.api_endpoint}/{url}' if self.api_endpoint not in url else url
        logging.debug(f'Performing DELETE request in "{url}" with params({params})')
        params = {} if params is None else params
        req = self.session.delete(url)
        self.update_csrf_token(req)
        headers = self.req_headers
        if content_type != headers['Content-type']:
            headers['Content-type'] = content_type
        resp = self.session.delete(url, params=params, headers=headers)

        if resp.status_code in (204, 200):
            # Success DELETE request
            logging.info(f'Successfully performed DELETE request on endpoint {url}.')
            return

        logging.error(f'Could not DELETE on endpoint: {url} With params: {params}')
        logging.error(f'Statuscode: {resp.status_code}')
        exception = requests.exceptions.RequestException(f'\nStatuscode: {resp.status_code}\n'
                                                         f'\nWith params: {params}\nOn endpoint:\n\t{url}')
        raise exception

    def add_object(self, obj: DSpaceObject) -> DSpaceObject | Collection | Item | Community:
        """
        Creates a new object in the DSpace Instance.

        :param obj: The object to create.
        :return: The newly created object.
        """
        if not self.authenticated:
            logging.critical('Could not add object, authentication required!')
            raise ConnectionRefusedError('Authentication needed.')
        params = {}
        match obj.get_dspace_object_type():
            case 'Item':
                obj: Item
                add_url = f'{self.api_endpoint}/core/items'
                params = {'owningCollection': obj.get_owning_collection().uuid}
            case 'Community':
                obj: Community
                if obj.parent_community is None:
                    add_url = f'{self.api_endpoint}/core/communities'
                else:
                    add_url = f'{self.api_endpoint}/core/communities'
                    params = {'parent': obj.parent_community.uuid}
            case 'Collection':
                obj: Collection
                add_url = f'{self.api_endpoint}/core/collections'
                params = {'parent': obj.community.uuid}
            case _:
                raise ValueError(f'Object type {obj.get_dspace_object_type()} is not allowed as a parameter!')
        obj_json = object_to_json(obj)

        return json_to_object(self.post_api(add_url, json_data=obj_json, params=params))

    def add_bundle(self, bundle: Bundle, item_uuid: str) -> Bundle:
        """
        Creates a new bundle based on a given bundle Object in DSpace and returns the created object.
        :param bundle: The bundle object to create.
        :param item_uuid: The uuid_of the item to create the bundle for.
        :return: The newly created object returned from DSpace.
        """
        if not self.authenticated:
            logging.critical('Could not add object, authentication required!')
            raise ConnectionRefusedError('Authentication needed.')
        params = {}
        add_url = f'{self.api_endpoint}/core/items/{item_uuid}/bundles'
        obj_json = {'name': bundle.name}
        if bundle.description != '':
            obj_json['metadata'] = {'dc.description': [{'value': bundle.description}]}
        resp = self.post_api(add_url, json_data=obj_json, params=params)
        uuid = resp['uuid']
        name = resp['name']
        return Bundle(name=name, uuid=uuid)

    def add_bitstream(self, bitstream: Bitstream, bundle: Bundle) -> str:
        """
        Creates a new bitstream in a given dspace bundle.
        :param bitstream: The bitstream to upload.
        :param bundle: The bundle to upload the item in.
        :return: The uuid of the newly created bitstream.
        """
        if not self.authenticated:
            logging.critical('Could not add object, authentication required!')
            raise ConnectionRefusedError('Authentication needed.')
        add_url = f'{self.api_endpoint}/core/bundles/{bundle.uuid}/bitstreams'
        obj_json = {'name': bitstream.file_name, 'metadata': {'dc.title': [{'value': bitstream.file_name}],
                                                              'dc.description': [{'value': bitstream.description}]},
                    'bundleName': bundle.name}
        if isinstance(bitstream, IIIFBitstream):
            bitstream: IIIFBitstream
            obj_json['metadata']['iiif.label'] = [{'value': bitstream.iiif['label']}]
            obj_json['metadata']['iiif.toc'] = [{'value': bitstream.iiif['toc']}]
            obj_json['metadata']['iiif.image.width'] = [{'value': bitstream.iiif['w']}]
            obj_json['metadata']['iiif.image.height'] = [{'value': bitstream.iiif['h']}]
        logging.debug(f'Adding bitstream: {obj_json}')
        bitstream_file = bitstream.get_bitstream_file()
        data_file = {'file': (bitstream.file_name, bitstream_file)} if bitstream_file is not None else None
        req = self.session.post(add_url)
        self.update_csrf_token(req)
        headers = self.session.headers
        headers.update({'Content-Encoding': 'gzip', 'User-Agent': self.req_headers['User-Agent']})
        req = requests.Request('POST', add_url,
                               data={'properties': json.dumps(obj_json) + ';type=application/json'}, headers=headers,
                               files=data_file)
        resp = self.session.send(self.session.prepare_request(req))
        try:
            uuid = resp.json()['uuid']
            logging.info(f'Successfully added bitstream with uuid "{uuid}"')
            return uuid
        except KeyError as e:
            logging.error(f'Problem with adding bitstream:\n{resp}\n\t{resp.headers}')
            raise e

    def add_relationship(self, relation: Relation) -> dict:
        """
        Creates a new relationship between to items based on the information in the Relation object.

        :param relation: The relation to create.
        """
        if not self.authenticated:
            logging.critical('Could not add object, authentication required!')
            raise ConnectionRefusedError('Authentication needed.')
        if relation.relation_type is None:
            logging.info('No relation type specified, trying to find relation-type via the rest endpoint.')
            left_item_type = relation.items[0].get_entity_type()
            rels = self.get_relations_by_type(left_item_type)
            rels = list(filter(lambda x: x.relation_key == relation.relation_key, rels))
            if len(rels) != 1:
                if len(rels) > 1:
                    logging.critical('Something went wrong with on the rest-endpoint: found more than one relation with'
                                     f' the name {relation.relation_key}')
                else:
                    logging.error(f'Didn\'t find relation with name {relation.relation_key}')
            else:
                relation.relation_type = rels[0].relation_type
                logging.debug(f'Found relationtype "{relation.relation_type}" for the name "{relation.relation_key}"')
        add_url = f'{self.api_endpoint}/core/relationships?relationshipType={relation.relation_type}'
        if relation.items[0] is None or relation.items[1] is None:
            logging.error(f'Could not create Relation because of missing item information in relation: {relation}')
            raise ValueError(f'Could not create Relation because of missing item information in relation: {relation}')
        uuid_1 = relation.items[0].uuid
        uuid_2 = relation.items[1].uuid
        if uuid_1 == '' or uuid_2 == '':
            logging.error(f'Relation via RestAPI can only be created by using item-uuids, but found: {relation}')
            raise ValueError(f'Relation via RestAPI can only be created by using item-uuids, but found: {relation}')
        req = self.session.post(add_url)
        self.update_csrf_token(req)
        item_url = f'{self.api_endpoint}/core/items'
        headers = self.session.headers
        headers.update({'Content-Type': 'text/uri-list', 'User-Agent': self.req_headers['User-Agent']})
        resp = self.session.post(add_url, f'{item_url}/{uuid_1} \n {item_url}/{uuid_2}', headers=headers)

        if resp.status_code in (201, 200):
            # Success post request
            logging.info(f'Created relationship: {relation}')
            return resp.json()

        raise requests.exceptions.RequestException(f'{resp.status_code}: Could not post relation: \n{relation}\n'
                                                   f'Got headers: {resp.headers}')

    def add_community(self, community: Community | DSpaceObject, create_tree: bool = False) -> Community:
        """
        Creates a new community in the DSpace instance and its owning community if create_tree is True.

        :param community: The collection object to create in DSpace.
        :param create_tree: If the owning communities shall be created as well.
        :return: Returns the newly created Community.
        """
        parent_community = community.parent_community
        if parent_community is not None and parent_community.uuid == '' and create_tree:
            community.parent_community = self.add_community(parent_community, create_tree)
        logging.debug(f'Adding community: {community}')
        return self.add_object(community)

    def add_collection(self, collection: Collection, create_tree: bool = False) -> Collection:
        """
        Creates a new collection in the DSpace instance and its owning communities if create_tree is True.

        :param collection: The collection object to create in DSpace.
        :param create_tree: If the owning communities shall be created as well.
        :return: Returns the newly created Collection.
        """
        community = collection.community
        if community.uuid == '' and create_tree:
            collection.community = self.add_community(community, create_tree)
        logging.debug(f'Adding collection: {collection}')
        return self.add_object(collection)

    def add_item(self, item: Item, create_tree: bool = False) -> Item:
        """
        Adds an item object to DSpace including files and relations. Based on the add_object method.

        :param item: The item to push into DSpace.
        :param create_tree: Creates the owning collections and communities above this item if not yet existing.
        :return: An item object including the new uuid.
        """
        bitstreams = item.contents
        collection_list = item.collections
        if create_tree:
            if len(collection_list) > 0 and collection_list[0].uuid == '':
                col: Collection = self.add_collection(collection_list[0], create_tree)
                collection_list[0] = col
        elif len(collection_list) == 0:
            raise ValueError('Can not push an Item into the restAPI without information about the owning collections.')
        elif len(collection_list) > 0 and collection_list[0].get_identifier() is None:
            raise ValueError('Can not push an Item into the restAPI without a owning collections. Set create_tree to'
                             'True or provide an identifier of the owning collection.')
        else:
            for c in collection_list:
                if c.uuid == '' and c.handle != '':
                    logging.debug(f'Could not find uuid for collection with handle "{c.handle}".'
                                  'Retrieving uuid from api.')
                    c.uuid = self.get_dso(identifier=c.handle).uuid
        dso = self.add_object(item)
        bundles = {i.name: i for i in [self.add_bundle(b, dso.uuid) for b in item.get_bundles()]}
        for b in bitstreams:
            self.add_bitstream(b, bundles[b.bundle.name])
        item.uuid = dso.uuid
        relations = item.relations if item.is_entity() else []
        if len(relations) > 0:
            relation_types = {r.relation_key: r.relation_type for r in self.get_relations_by_type(
                item.get_entity_type())}
            try:
                relations = list(map(lambda x: Relation(x.relation_key, x.items, relation_types[x.relation_key]),
                                     relations))
            except KeyError as e:
                logging.error(f'Could not find relation in the list: {relation_types}')
                raise e
        for r in relations:
            self.add_relationship(r)
        logging.debug(f'Created item {item}')
        return item

    def move_item(self, item: Item, new_collection: Collection | str):
        """
        Moves an item to a new collection (completely replacing the old one).

        :param item: The item to move.
        :param new_collection: The new collection to put the item. This can eather be a Collection object or a string
            containing the uuid of the collection.
        """
        endpoint = f'core/items/{item.uuid}/owningCollection'
        new_collection = new_collection.uuid if isinstance(new_collection, Collection) else new_collection
        logging.info(f'Moving item with uuid {item.uuid} into new owning collection {new_collection}.')
        self.put_api(endpoint, f'{self.api_endpoint}/core/collections/{new_collection}',
                     content_type='text/uri-list')

    def get_dso(self, uuid: str = '', endpoint: str = '',
                identifier: str = None) -> DSpaceObject | Item | Collection | Community:
        """
        Retrieves a DSpace object from the api based on its uuid and the endpoint information.

        :param uuid: The uuid of the object.
        :param endpoint: The endpoint string. Must be one of ('items', 'collections', 'communities')
        :param identifier: An optional other identifier to retrieve a DSpace Object. Can be used instead of uuid. Must
            be a handle or doi.
        :return: Returns a DSpace object
        """
        if identifier is None:
            if uuid == '' or not isinstance(uuid, str):
                raise ValueError(f'If no other identifier is used, the uuid must be provided. "{uuid}" is not correct.')
            params = {}
            url = f'core/{endpoint}/{uuid}'
            if endpoint not in ('items', 'collections', 'communities'):
                raise ValueError(f"The endpoint '{endpoint}' does not exist. endpoint must be one of"
                                 "('items', 'collections', 'communities')")
        else:
            url = 'pid/find'
            params = {'id': identifier}
        try:
            obj = json_to_object(self.get_api(url, params))
            logging.debug(f'Retrieved DSpaceObject: {obj}')
        except TypeError:
            obj = None
            logging.warning('The object could not be found!')
        return obj

    def get_paginated_objects(self, endpoint: str, object_key: str, query_params: dict = None, page: int = -1,
                              size: int = 20) -> list[dict]:
        """
        Retrieves a paginated list of objects from the remote dspace endpoint and returns them as a list.

        :param endpoint: The endpoint to retrieve the objects from.
        :param object_key: The dict key to get the object list from the json-response. For example "bundles" or
            "bitstreams"
        :param query_params: Additional query parameters to add to the request.
        :param page: The page number to retrieve. Must be set to -1 to retrieve all pages. Default -1.
        :param size: The page size, aka the number of objects per page.
        :return: The list of retrieved objects.
        """
        query_params = {} if query_params is None else query_params
        if size > -1:
            query_params.update({'size': size})
        if page > -1:
            query_params.update({'page': page})
        endpoint_json = self.get_api(endpoint, params=query_params)
        if 'discover/search/objects' in endpoint:
            endpoint_json = endpoint_json['_embedded']['searchResult']
        try:
            object_list = endpoint_json["_embedded"][object_key]
        except KeyError as e:
            logging.error(f'Problems with parsing paginated object list. A Key-Error occurred.\n{endpoint_json}')
            raise e
        if page == -1:
            page_info = endpoint_json["page"]
            for p in range(1, page_info['totalPages']):
                object_list += self.get_paginated_objects(endpoint, object_key, query_params, p, size)
        return object_list

    def get_item_bitstreams(self, item_uuid: str) -> list[Bitstream]:
        """
        Retrieves the bitstreams connected to a DSpace Object. And returns them as a list.

        :param item_uuid: The uuid of the item to retrieve the bitstreams from.
        :return: A list of Bitstream objects.
        """
        bitstreams = []
        bundles = self.get_item_bundles(item_uuid)
        for b in bundles:
            bitstreams += self.get_bitstreams_in_bundle(b).bitstreams
        return bitstreams

    def get_bitstreams_in_bundle(self, bundle: Bundle) -> Bundle:
        """
        Retrieves all bitstreams in a given bundle by the bundle uuid.

        :param bundle: The bundle object to retrieve the bitstreams to.
        :return: The updated bundle object containing the bitstreams associated.
        """
        bitstream_link = f"/core/bundles/{bundle.uuid}/bitstreams"
        logging.debug(f'Retrieving bitstreams for bundle({bundle.name}) with uuid: {bundle.uuid}')
        for o in self.get_paginated_objects(bitstream_link, 'bitstreams'):
            description = o['metadata']['dc.description'][0]['value'] if ('dc.description'
                                                                          in o['metadata'].keys()) else ''
            bitstream = Bitstream(o['name'], o['_links']['content']['href'],
                                  bundle=bundle, uuid=o['uuid'])
            bitstream.add_description(description)
            bundle.add_bitstream(bitstream)
            logging.debug(f'Retrieved bitstreams: {bitstream}')
        logging.debug(f'Retrieved {len(bundle.bitstreams)} bitstreams.')
        return bundle

    def get_item_bundles(self, item_uuid: str) -> list[Bundle]:
        """
        Retrieves the bundles connected to a DSpaceObject and returns them as list.

        :param item_uuid: The uuid of the item to retrieve the bundles from.
        :return: The list of Bundle objects.
        """
        bundle_json = self.get_paginated_objects(f'/core/items/{item_uuid}/bundles', 'bundles')
        return [Bundle(b["name"],
                       b['dc.description'][0]['value'] if 'dc.description' in b["metadata"].keys() else '',
                       b['uuid']) for b in bundle_json]

    def get_relations_by_type(self, entity_type: str) -> list[Relation]:
        """
        Parses the REST API and returns a list of relationships, which have the given entity on the left or right side.

        :param entity_type: The entity_type to look for.
        :return: Return s a list of relations.
        """
        add_url = f'/core/relationshiptypes/search/byEntityType'
        params = {'type': entity_type}
        rel_list = []
        relations = self.get_paginated_objects(add_url, 'relationshiptypes', params)
        for r in relations:
            rel_list.append(Relation(r['leftwardType'], relation_type=r['id']))
            rel_list.append(Relation(r['rightwardType'], relation_type=r['id']))
            logging.debug(f'Got relation {r} from RestAPI')
        return rel_list

    def get_item_relationships(self, item_uuid: str) -> list[Relation]:
        """
        Retrieves a list of relationships of DSpace entity from the api.

        :param item_uuid: The uuid of the item to retrieve the relationships for.
        :return: A list of relation objects.
        """
        url = f'/core/items/{item_uuid}/relationships'

        rel_list = self.get_paginated_objects(url, 'relationships')
        relations = []
        for r in rel_list:
            left_item_uuid = r['_links']['leftItem']['href'].split('/')[-1]
            right_item_uuid = r['_links']['rightItem']['href'].split('/')[-1]
            direction = 'leftwardType' if item_uuid == right_item_uuid else 'rightwardType'
            # Retrieve the type information:
            type_req = self.session.get(r['_links']['relationshipType']['href'])
            rel_key = type_req.json()[direction]
            rel_type = type_req.json()['id']
            # Set the correct item order.
            try:
                left_item = self.get_item(left_item_uuid, False)
                right_item = self.get_item(right_item_uuid, False)
                items = (left_item, right_item) if direction == 'rightwardType' else (right_item, left_item)
                relation = Relation(rel_key, items, rel_type)
                relations.append(relation)
                logging.debug(f'Added relation {relation} to Item.')
            except requests.exceptions.RequestException:
                logging.warning(f'Could not retrieve relationship({rel_key}) between {left_item_uuid} and'
                                f' {right_item_uuid}')
        return relations

    def get_item_collections(self, item_uuid: str) -> list[Collection]:
        """
        Retrieves a list of collections from the REST-API based on the uuid of an item. The first will be the owning
        collection.

        :param item_uuid: The uuid of the item.
        """
        url = f'core/items/{item_uuid}/owningCollection'
        get_result = self.get_api(url)
        if get_result is None:
            logging.warning(f'Problems with getting owning Collection for item with uuid "{item_uuid}"')
            return []
        owning_collection = json_to_object(get_result)
        mapped_collections = self.get_paginated_objects(f'core/items/{item_uuid}/mappedCollections',
                                                        'mappedCollections')
        return [owning_collection] + list(filter(lambda x: x is not None,
                                                 [json_to_object(m) for m in mapped_collections]))

    def get_parent_community(self, dso: Collection | Community) -> Community | None:
        """
        Retrieves the parent community of a given collection or Community.

        :param dso: The object to get the parent community from. Must be either Collection or Community
        """
        url = f'core/{"collections" if isinstance(dso, Collection) else "communities"}/{dso.uuid}/parentCommunity'
        try:
            get_result = self.get_api(url)
        except requests.exceptions.RequestException:
            return None
        if get_result is None:
            logging.warning(f'Problems with getting parent communities for DSpaceObject {dso}')
            return None
        owning_community = json_to_object(get_result)
        return owning_community

    def get_item(self, uuid: str = '', get_related: bool = True, get_bitstreams: bool = True,
                 pre_downloaded_item: Item = None, identifier: str = None) -> Item | None:
        """
        Retrieves a DSpace-Item by its uuid (or other identifieres) from the API.

        :param uuid: The uuid of the item to get.
        :param get_related: If true, also retrieves related items from the API.
        :param get_bitstreams: If true, also retrieves bitstreams of the item from the API.
        :param pre_downloaded_item: If a pre downloaded item is provided (aka blank dso), then only additional
            information such as relationships, owning collection, bundles and bitstreams will be provided.
        :param identifier: The identifier of the item to get. Can be used to retrieve items with ids other than uuid.
            Must be doi or handle
        :return: An object of the class Item.
        """
        dso = self.get_dso(uuid, 'items', identifier) if pre_downloaded_item is None else pre_downloaded_item
        if dso is None:
            logging.warning(f'The item with uuid "{uuid}" could not be found.')
            return None
        if get_related:
            dso.relations = self.get_item_relationships(dso.uuid)
        if get_bitstreams:
            dso.bundles = [self.get_bitstreams_in_bundle(b) for b in self.get_item_bundles(dso.uuid)]
            for b in dso.bundles:
                dso.contents += b.bitstreams

        dso.collections = self.get_item_collections(dso.uuid)
        logging.debug(f'Successfully retrieved item {dso} from endpoint.')
        return dso

    def get_community(self, uuid) -> Community | None:
        """
        Retrieves a DSpace-Community object from the API.

        :param uuid: The UUID of the community to get.
        """
        dso = self.get_dso(uuid, 'communities')
        dso: Community
        dso.parent_community = self.get_parent_community(dso)
        logging.debug(f'Successfully retrieved community {dso} from endpoint.')
        return dso

    def get_collection(self, uuid) -> Collection | None:
        """
        Retrieves a DSpace-Community object from the API.

        :param uuid: The UUID of the community to get.
        """
        dso = self.get_dso(uuid, 'collections')
        dso: Collection
        dso.community = self.get_parent_community(dso)
        logging.debug(f'Successfully retrieved collection {dso} from endpoint.')
        return dso

    def get_bundle(self, uuid: str) -> Bundle | None:
        """
        Retrieves a DSpace-Bundel object from the API.

        :param uuid: The UUID of the bundle to get.
        :return: The bundle found.
        """
        bundle_json = self.get_api(f'core/bundles/{uuid}')
        bundle: Bundle = Bundle(bundle_json['name'], uuid=bundle_json['uuid'])
        bundle = self.get_bitstreams_in_bundle(bundle)
        logging.debug(f'Successfully retrieved bundle {bundle} including {len(bundle.bitstreams)} bitstreams from'
                      f'endpoint.')
        return bundle

    def get_items_in_scope(self, scope_uuid: str, query: str = '', size: int = -1, page: int = -1,
                           full_item: bool = False) -> list[Item]:
        """
        Returns a list of DSpace items in a given collection or community. Can be further reduced by query parameter.

        :param scope_uuid: The uuid of the collection to retrieve the items from.
        :param query: Additional query parameters for the request.
        :param size: The number of objects per page. Use -1 to select the default.
        :param page: The page to retrieve if a paginated list is returned. Use -1 to retrieve all.
        :param full_item: If the full item information should be downloaded (Including relationships, bundles and
            bitstreams. This can be slower due to additional api calls).
        :return: A list of Item objects.
        """
        query_params = {'scope': scope_uuid}
        if query != '':
            query_params.update({'query': query})
        if page > -1:
            query_params.update({'page': page})

        search_req = self.get_api('discover/search/objects', query_params)
        json_res = search_req['_embedded']['searchResult']
        try:
            item_list = [json_to_object(i['_embedded']['indexableObject']) for i in json_res['_embedded']['objects']]
            item_list = list(filter(lambda x: isinstance(x, Item), item_list))
            if full_item:
                item_list = [self.get_item(i.uuid, True, True, i) for i in item_list]
        except KeyError as e:
            logging.error(f'Problem with parsing the following request anser:\n{json_res}')
            raise e

        if page == -1:
            number_pages = json_res['page']['totalPages']
            logging.debug(f'Parsing {number_pages} with items.')
            for n in range(1, number_pages):
                item_list += self.get_items_in_scope(scope_uuid, query, size, n, full_item)

        return list(filter(lambda x: x is not None, item_list))

    def search_items(self, query_params: dict = None, size: int = 20, full_item: bool = False,
                     get_bitstreams: bool = False) -> list[DSpaceObject]:
        """
        Search items via rest-API using solr-base query parameters. Uses the endpoint /discover/search/objects. If no
        query_params are provided, the whole repository will be retrieved.

        :param query_params: A dictionary with query parameters to filter the search results.
        :param size: The number of objects to retrieve per page.
        :param full_item: If the full items (including relations and bitstreams) shall be downloaded or not.
            Default false.
        :param get_bitstreams: If true, also the bitstreams of the item will be connected. Not needed if full_item is
            True. Default: False.
        :return: The list of found DSpace objects.
        """
        object_list = self.get_paginated_objects('/discover/search/objects', 'objects', query_params,
                                                 size=size)
        dspace_objects = [json_to_object(obj['_embedded']['indexableObject']) for obj in object_list]
        if not full_item and not get_bitstreams:
            logging.info(f'Found {len(dspace_objects)} DSpace Objects.')
            return dspace_objects

        for o in dspace_objects:
            if o.get_dspace_object_type() == 'Item':
                o: Item
                if full_item:
                    o.collections = self.get_item_collections(o.uuid)
                    o.relations = self.get_item_relationships(o.uuid)
                o.contents = self.get_item_bitstreams(o.uuid)
            if full_item:
                if o.get_dspace_object_type() == 'Collection':
                    o: Collection
                    o.community = self.get_parent_community(o)
                if o.get_dspace_object_type() == 'Community':
                    o: Community
                    o.parent_community = self.get_parent_community(o)
        logging.info(f'Found {len(dspace_objects)} DSpace Objects.')
        return dspace_objects

    def get_all_items(self, page_size: int = 20, full_item: bool = False,
                      get_bitstreams: bool = False) -> list[DSpaceObject]:
        """
        Retrieves all Items from the REST-API. This method simply refers to the `search_items()` method without using
        query_params, thus getting all items.

        :param page_size: The number of objects to retrieve per page.
        :param full_item: Whether the full_item including related items and bitstreams shall be downloaded.
        :param get_bitstreams: If true, also the bitstreams of the item will be connected. Not needed if full_item is
            True. Default: False.
        :return: A list of all found DSpaceObjects
        """
        return self.search_items(None, page_size, full_item, get_bitstreams)

    def get_metadata_field(self, schema: str = '', element: str = '', qualifier: str = '',
                           field_id: int = -1) -> list[dict]:
        """
        Checks if given metadata field exists in the DSpace instance. Returns one or more found metadata fields in a
        list of dict.

        :param schema: The schema of the field, if empty this field won't be taken in account for the search request.
        :param element: The element of the field, if empty this field won't be taken in account for the search request.
        :param qualifier: The qualifier of the field, if empty this field won't be taken in account for the search
            request.
        :param field_id: The exact metadata field id to look for. If the correct fields is already known.
        :return: A list of dictionaries in the following form: {id: <id>, element: <element>, qualifier: <qualifier>,
            scopeNote: <scopeNote>, schema: {id: <schema-id>, prefix: <prefix>, namespace: <namespace>}
        """

        def parse_json_resp(json_resp: dict) -> dict:
            """
            Parses the answer of the REST-API response and returns it into the wanted format.
            """
            schema_resp = json_resp['_embedded']['schema']
            return {'id': json_resp['id'],
                    'element': json_resp['element'],
                    'qualifier': json_resp['qualifier'],
                    'scopeNote': json_resp['scopeNote'],
                    'schema': {'id': schema_resp['id'], 'prefix': schema_resp['prefix'],
                               'namespace': schema_resp['namespace']}
                    }

        url = 'core/metadatafields/'
        if field_id > -1:
            url += f'/{field_id}'
            json_get = self.get_api(url)
            return [] if json_get is None else [parse_json_resp(json_get)]

        url += 'search/byFieldName'
        params = {}
        if schema != '':
            params.update({'schema': schema})
        if element != '':
            params.update({'element': element})
        if qualifier != '':
            params.update({'qualifier': qualifier})
        json_get = self.get_api(url, params)
        if json_get is None:
            return []
        pages = json_get['page']['totalPages']
        current_page = json_get['page']['number'] + 1
        results = [json_get]
        for p in range(current_page, pages):
            params['page'] = p
            results.append(self.get_api(url, params))
        field_objects = []
        for r in results:
            field_objects += [parse_json_resp(i) for i in r['_embedded']['metadatafields']]
            logging.debug(f'Found metadata field {field_objects[-1]}')
        logging.info(f'Found {len(field_objects)} metadata fields.')
        return field_objects

    def update_metadata(self, metadata: dict[str, (list[dict] | dict[str, dict])], object_uuid: str, obj_type: str,
                        operation: str, position: int = -1) -> DSpaceObject:
        """
        Update a new metadata value information to a DSpace object, identified by its uuid.

        :param metadata: A list of metadata to update as a MetaData object or dict object in the REST form, aka
            {<tag> : [{"value": <value>, "language": <language>...}]}. May also contain position information. For
            "remove"- operation the form must be {<tag>: [{postion: <position>}] | []}
        :param object_uuid: The uuid of the object to add the metadata to.
        :param obj_type: The type of DSpace object. Must be one of item, collection or community
        :param operation: The selected update operation. Must be one off (add, replace, remove).
        :param position: The position of the metadata value to add. Only possible if metadata is of type dict[dict[]]
        :return: The updated DSpace object.
        :raises ValueError: If a not existing objectType is used or wrong operation type.
        """
        if obj_type not in ('item', 'collection', 'community'):
            logging.error(f'Wrong object type information "{obj_type}" must be one of item, collection or community')
            raise ValueError(f'Wrong object type information "{obj_type}" must be one of item, collection or community')
        if operation not in ('add', 'replace', 'remove'):
            logging.error(f'Wrong update operation "{operation}" must be one off (add, replace, remove).')
            raise ValueError(f'Wrong update operation "{operation}" must be one off (add, replace, remove).')

        patch_json = []
        # Form: [{"op": "<operation>",
        #            "path": "/metadata/<tag>",
        #            "value": [{"value": <value>, "language": <language>}]}]

        if operation in ('add', 'replace'):
            for k in metadata.keys():
                values = metadata[k]
                rank = ''
                if isinstance(metadata[k], dict):
                    metadata[k]: dict
                    rank = f'/{metadata[k]["position"]}' if "position" in metadata[k].keys() else ''
                    rank = f'/{position}' if rank == '' and str(position) != '-1' else rank
                    values = {"value": metadata[k]["value"]}
                    if "language" in metadata[k].keys():
                        values.update({"language": metadata[k]["language"]})
                patch_json.append({"op": operation, "path": f"/metadata/{k}" + rank, "value": values})
        else:
            for k in metadata.keys():
                rank = [i['position'] for i in metadata[k]]
                if len(rank) > 0:
                    patch_json += [{"op": operation, "path": f"/metadata/{k}/{r}"} for r in rank]
                else:
                    patch_json.append({"op": operation, "path": f"/metadata/{k}" +
                                                                (f'/{position}' if str(position) != '-1' else '')})

        url = 'core/' + (f'{obj_type}s' if obj_type in ('item', 'collection') else 'communities')
        json_resp = self.patch_api(f'{url}/{object_uuid}', patch_json)
        return json_to_object(json_resp)

    def add_metadata(self, metadata: MetaData | dict[str, list[dict]], object_uuid: str,
                     obj_type: str, position_end: bool = False) -> DSpaceObject:
        """
        Add a new metadata value information to a DSpace object, identified by its uuid.

        :param metadata: A list of metadata to update as a MetaData object or dict object in the REST form, aka
            {<tag> : [{"value": <value>, "language": <language>...}]}
        :param object_uuid: The uuid of the object to add the metadata to.
        :param obj_type: The type of DSpace object. Must be one of item, collection or community.
        :param position_end: Whether the new metadata field should be placed at the end of the existing metadata.
        :return: The updated DSpace object.
        :raises ValueError: If a not existing objectType is used.
        """

        if isinstance(metadata, MetaData):
            metadata = {key: [dict(v) for v in metadata[key]] for key in metadata.keys()}
        else:
            # Checks if there is only one metadata key with only one value.
            if len(metadata.keys()) == 1 and position_end and len(metadata[list(metadata.keys())[0]]) == 1:
                metadata = {list(metadata.keys())[0]: metadata[list(metadata.keys())[0]][0]}
        return self.update_metadata(metadata, object_uuid, obj_type, 'add',
                                    position='-' if position_end else -1)

    def replace_metadata(self, metadata: MetaData | dict[str, list[dict] | dict], object_uuid: str,
                         obj_type: str, position: int = -1) -> DSpaceObject:
        """
        Add a new metadata value information to a DSpace object, identified by its uuid.

        :param metadata: A list of metadata to update as a MetaData object or dict object in the REST form, aka
            {<tag> : [{"value": <value>, "language": <language>...}]}
        :param object_uuid: The uuid of the object to add the metadata to.
        :param obj_type: The type of DSpace object. Must be one of item, collection or community.
        :param position: The position of the metadata value to replace.
        :return: The updated DSpace object.
        :raises ValueError: If a not existing objectType is used.
        """
        if isinstance(metadata, MetaData):
            metadata: MetaData
            patch_data = {key: [dict(v) for v in metadata[key]] for key in metadata.keys()}
        else:
            metadata: dict[str, list[dict] | dict]
            # Check if position argument is not used correctly
            if len(metadata.keys()) > 1 and str(position) != '-1':
                logging.warning('Could not set same position metadata for more than one metadata tag.')
                raise Warning('Could not set same position metadata for more than one metadata tag.')
            if (len(metadata.keys()) == 1 and isinstance(metadata[list(metadata.keys())[0]], list)
                    and str(position) != '-1'):
                logging.warning('Could not use one position argument for more than one metadata-value.')
                raise Warning('Could not use one position argument for more than one metadata-value.')

            patch_data = {k: (metadata[k][0] if (isinstance(metadata[k], list)
                                                 and len(metadata[k]) == 1) else metadata[k])for k in metadata.keys()}

        return self.update_metadata(patch_data, object_uuid, obj_type, 'replace', position=position)

    # Delete section. Be carefully, when using it!

    def delete_metadata(self, tag: str | list[str], object_uuid: str,
                        obj_type: str, position: int | str = -1) -> DSpaceObject:
        """
        Deletes a specific metadata-field or value of a DSpace Item. Can delete a list of fields as well as only one.

        :param tag: A tag or list of tags wich shall be deleted.
        :param object_uuid: The uuid of the DSpace Item to delete the metadata from.
        :param position: The position of the metadata value to delete. Can only be used, if only one tag is provided.
        :param obj_type: The type of DSpace object. Must be one of item, collection or community.
        :return: The updated DSpace object.
        :raises ValueError: If a not existing objectType is used. Or a position is given, when tag is a list.
        """
        if str(position) != '-1' and isinstance(tag, list):
            raise ValueError('Can not use position parameter if more than one tag is provided.')
        tag = [tag] if not isinstance(tag, list) else tag
        return self.update_metadata({t: [] for t in tag}, object_uuid, obj_type, operation='remove', position=position)

    def delete_bitstream(self, bitstream_uuid: str | list[str]):
        """
        Permanently removes a bitstream of a list of bitstreams from the repository. Handle be carefully when using the
        method, there won't be a confirmation step.

        :param bitstream_uuid: The uuid of the bitstream to delete
        """
        bitstream_uuid = [bitstream_uuid] if isinstance(bitstream_uuid, str) else bitstream_uuid
        patch_call = [{"op": "remove", "path": f"/bitstreams/{uuid}"} for uuid in bitstream_uuid]
        self.patch_api("core/bitstreams", patch_call)
        logging.info(f'Successfully deleted bitstream with uuid "{bitstream_uuid}".')

    def delete_bundles(self, bundle_uuid: str | list[str], include_bitstreams: bool = False):
        """
        Permanently removes a bundle or a list of bundles from the repository. Handle carefully when using the method,
        there won't be a confirmation step. Can not delete a bundle with bitstreams still included, unless
        include_bitstreams is set to true.

        :param bundle_uuid: The uuid or list of uuids of the bundles to be deleted.
        :param include_bitstreams: Default: False. WARNING! If this is set to true all bitsreams in the bundle will be
            deleted as well.
        """
        bundles = bundle_uuid if isinstance(bundle_uuid, list) else [bundle_uuid]
        for b in bundles:
            if not include_bitstreams:
                bundle = self.get_bitstreams_in_bundle(Bundle(uuid=b))
                if len(bundle.bitstreams) > 0:
                    logging.error(f'Could not delete bundle with uuid "{b}" because there are still '
                                  f'{len(bundle.bitstreams)} bitstreams.')
                    continue
            self.delete_api(f'core/bundles/{b}')
            logging.info(f'Successfully deleted bundle with uuid "{b}"')



import unittest

from dspyce._testing.dspyceTest import DSpaceObjectTest
from dspyce._testing.statisticsTest import StatisticTest
from dspyce._testing.metadataTests import MetadataTest
from dspyce._testing.itemTests import ItemTest
from dspyce._testing.relationTest import RelationTest
from dspyce._testing.bitstreamTest import BitstreamsTest
from dspyce._testing.safTest import SAFTest


class InitTests(unittest.TestCase):
    """Dummy test class"""
    pass


if __name__ == '__main__':
    unittest.main()



# Relation.py
"""
This module contains the python class Relation representing the Relation between DSpace-Entities.
And the following function:
 - `export_relations(relations: list[Relation])` - Returns a list of relationships in a string.

Example:

    >from saf import export_relations
    >
    >export_relations([Relation('any_relation', '123456789/12'), Relation('different_relation', '123456789/13')])

    > relation.any_relation 123456789/12\nrelation.different_relation 123456789/13\n

"""
from dspyce import Item


class Relation:
    """
        The class Relation represents the relation between different DSPACE-entities. It stores the relation-type
        (relation_key) and id of the related Item.
    """
    relation_key: str
    relation_type: int
    items: tuple[Item, Item]

    def __init__(self, relation_key: str, related_items: tuple[Item, Item] = None, relation_type: int = None):
        """
            Creates a new object of the class relation, which represents exactly
            one DSpace-Relation

            :param relation_key: The relation-type, aka the name.
            :param related_items: The two related entities. Attention: The order will always be preserved.
            :param relation_type: A DSpace specific relation ID if existing.
        """
        self.relation_key = relation_key
        self.items = related_items if related_items is not None else (None, None)
        self.relation_type = relation_type

    def __str__(self):
        id_1 = self.items[0].get_identifier() if self.items[0] is not None else 'None'
        id_2 = self.items[1].get_identifier() if self.items[1] is not None else 'None'
        return f'{id_1}:relation.{self.relation_key}:{id_2}'

    def __eq__(self, other):
        return self.relation_key == other.relation_key and self.relation_type == other.relation_type

    def set_relation_type(self, relation_type: int):
        """
        Set a value for the relation_type variable.
        :param relation_type: The new relation_id value.
        """
        self.relation_type = relation_type



import unittest
from dspyce.statistics.retrieve import get_point_views


class StatisticTest(unittest.TestCase):
    """
    Tests for statistics package.
    """
    def test_get_point_views(self):
        """
        Tests get_point_views function from statistics package.
        """
        self.assertEqual(get_point_views(
            {
                "id": "8ad84c46-bdf1-4558-9b90-5c43ff396980",
                "label": "Journals",
                "values": {
                    "views": 151
                },
                "type": "collection"
            }), 151)
        self.assertIsNone(get_point_views({"id": "8ad84c46-bdf1-4558-9b90-5c43ff396980", "label": "Journals"}))
        self.assertRaises(TypeError, get_point_views, 'Test')


if __name__ == '__main__':
    unittest.main()



import unittest

import dspyce as ds
from dspyce.bitstreams import Bundle


class ItemTest(unittest.TestCase):
    item: ds.Item = ds.Item('xyz', '123456789/1', 'test', ds.Collection('abc'))

    def test_entity(self):
        """
        Tests all entity related methods.
        """
        self.item.enable_entity('Publication')
        self.assertTrue(self.item.is_entity())
        self.assertEqual('Publication', self.item.get_metadata_values('dspace.entity.type')[0])
        self.assertEqual('Publication', self.item.get_entity_type())
        self.item.add_relation('isAuthorOfPublication', 'xyz-uuid')
        self.assertRaises(TypeError, ds.Item('').add_relation, 'isContributionOfPublication', 'xyz-uuid')
        self.assertEqual(ds.Item(uuid='xyz-uuid'), self.item.get_related()[0])

    def test_object_type(self):
        """
        Tests object_type method.
        """
        self.assertEqual('Item', self.item.get_dspace_object_type())

    def test_bitstreams(self):
        """
        Tests bitstream related methods.
        """
        self.item.add_content('TEST-FILE', '/test/path/to/file.txt', 'description')
        self.item.add_content('TEST-FILE-2', '/test/path/to/file.txt', 'description',
                              bundle=Bundle('TEST'))
        self.assertEqual([Bundle('ORIGINAL'), Bundle('TEST')], self.item.get_bundles())

    def test_collections(self):
        """
        Tests collection related methods.
        """
        self.assertEqual(ds.Collection('abc'), self.item.get_owning_collection())
        self.item.add_collection(ds.Collection('dfg'))
        self.assertEqual(ds.Collection('dfg'), self.item.collections[-1])


if __name__ == '__main__':
    unittest.main()



import os
import re
import requests


class Bitstream:
    """
        A class for managing bitstream files in the DSpace context.
    """
    file_name: str
    """The name of the file."""
    path: str
    """The path, where the file can be found."""
    description: str
    """A possible description of the bitstream."""
    permissions: list[dict[str, str]]
    """Permission which group shall have access to this file."""
    show: bool
    """If the file should be accessible for users or only provides information for the item import."""
    bundle: any
    """The bundle where to store the file. The default is set to the variable DEFAULT_BUNDLE."""
    primary: bool
    """If the bitstream shall be the primary bitstream for the item."""
    uuid: str
    """A uuid if the Bitstream already exists in a DSpace-Instance."""

    def __init__(self, name: str, path: str, bundle: any = None, uuid: str = None, primary: bool = False):
        """
        Creates a new Bitstream object.
        :param name: The name of the bitstream.
        :param path: The path, where the file is currently stored.
        :param bundle: The bundle, where the bitstream should be placed in. The default is ORIGINAL.
        :param uuid: The uuid of the bitstream if existing.
        :param primary: Primary is used to specify the primary bitstream.
        """
        self.file_name = name
        self.path = path
        self.path += '/' if len(self.path) > 0 and self.path[-1] != '/' else ''
        self.permissions = []
        self.description = ''
        self.bundle = bundle
        self.uuid = uuid
        self.primary = primary

    def __str__(self):
        """
        Provides all information about the DSpace-Content file.

        :return: A SAF-ready information string which can be used for the content-file.
        """
        export_name = self.file_name
        if self.bundle is not None:
            export_name += f'\tbundle:{self.bundle.name}'
        if self.description != '':
            export_name += f'\tdescription:{self.description}'
        if len(self.permissions) > 0:
            for p in self.permissions:
                export_name += f'\tpermissions:-{p["type"]} \'{p["group"]}\''
        if self.primary:
            export_name += '\tprimary:true'
        return export_name

    def add_description(self, description):
        """
            Creates a description to the content-file.

            :param description: String which provides the description.
        """
        self.description = description

    def add_permission(self, rw: str, group_name: str):
        """
            Add access information to the Bitstream.

            :param rw: Access-type r-read, w-write.
            :param group_name: Group to which the access will be provided.
        """
        if rw not in ('r', 'w'):
            raise ValueError(f'Permission type must be "r" or "w". Got {rw} instead!')
        self.permissions.append({'type': rw, 'group': group_name})

    def get_bitstream_file(self, timeout: int = 30) -> bytes:
        """
        Returns the actual file as a TextIOWrapper object.

        :param timeout: The connection timeout for reading bitstreams from remote resources.
        """
        if re.search(r'^http(s)?://', self.path):
            return requests.get(self.path, timeout=timeout).content
        with open(self.path + self.file_name, 'rb') as f:
            return f.read()

    def save_bitstream(self, path: str, timeout: int = 30):
        """
        Saves the current bitstream to the given path.

        :param path: The path where the bitstream file is to be saved.
        :param timeout: The connection timeout for reading bitstreams from remote resources.
        :raises FileExistsError: if the file already exists in the given path.
        """
        if self.file_name in os.listdir(path):
            raise FileExistsError(f'The file "{self.file_name}" already exists in {path}')
        file = self.get_bitstream_file(timeout)
        if isinstance(file, str):
            file = file.encode('utf-8')
        with open(f'{path}/{self.file_name}', 'wb') as f:
            f.write(file)

    def __eq__(self, other):
        """
        Checks if two content files are equal based on their name and path (and possible uuid).
        
        :param other: The other object to compare with.
        :raises TypeError: if the type of "other" is not Bitstream
        """
        if not isinstance(other, Bitstream):
            raise TypeError(f'Can not compare Bitstream with type({type(other)})')
        return (self.file_name == other.file_name and
                self.path == other.path and
                ((self.uuid is None or other.uuid is None) or self.uuid == other.uuid))



from .metadata import MetaData, MetaDataValue


class DSpaceObject:
    """
    The class DSpaceObject represents an Object in a DSpace repository, such as Items, Collections, Communities.
    """

    uuid: str
    """The uuid of the DSpaceObject"""
    name: str
    """The name of the DSpaceObject, if existing"""
    handle: str
    """The handle of the Object"""
    metadata: MetaData
    """The metadata provided for the object."""
    statistic_reports: dict
    """A dictionary of statistic report objects."""

    def __init__(self, uuid: str = '', handle: str = '', name: str = ''):
        """
        Creates a new object of the class DSpaceObject

        :param uuid: The uuid of the DSpaceObject. Default is ''
        :param handle: The handle of the DSpaceObject. Default is ''
        :param name: The name of the DSpaceObject, if existing.
        """
        self.uuid = uuid
        self.handle = handle
        self.name = name
        self.metadata = MetaData({})
        self.statistic_reports = {}

    def add_metadata(self, tag: str, value: str, language: str = None):
        """
        Creates a new metadata field with the given value.

        :param tag: The correct metadata tag. The string must use the format <schema>.<element>.<qualifier>.
        :param value: The value of the metadata-field.
        :param language: The language of the metadata field.

        :raises KeyError: If the metadata tag doesn't use the format <schema>.<element>.<qualifier>.'
        """
        self.metadata[tag] = MetaDataValue(value, language)

    def remove_metadata(self, tag: str, value: str = None):
        """
        Remove a specific metadata field from the DSpaceObject. Can either be all values for a field or ony a specific
        value based on the *value* parameter.

        :param tag: The correct metadata tag. The string must use the format <schema>.<element>.<qualifier>.
        :param value: The value of the metadata field to delete. Can be used, if only one value in a list of values
            should be deleted. If None, all values from the given tag will be deleted.
        """
        if value is None:
            self.metadata.pop(tag)
        else:
            self.metadata[tag] = list(filter(lambda x: x.value != value, self.metadata[tag]))

    def replace_metadata(self, tag: str, value: str, language: str = None):
        """
        Replaces a specific metadata field from the DSpaceObject. Replaces all values of a given tag.

        :param tag: The correct metadata tag. The string must use the format <schema>.<element>.<qualifier>.
        :param value: The value of the metadata field to add. Can be used, if only one value in a list of values
            should be deleted. If None, all values from the given tag will be deleted.
        :param language: The language of the metadata value to add.
        """
        self.remove_metadata(tag)
        self.add_metadata(tag, value, language)

    def get_dspace_object_type(self) -> str:
        """
        This Function serves mainly to be overwritten by subclasses to get the type of DSpaceObject.
        """
        pass

    def get_identifier(self) -> str | None:
        """
        Returns the identifier of this object. Preferably this will be the uuid, but if this does not exist, it uses
        the handle.
        :return: The identifier as a string.
        """
        if self.uuid != '':
            return self.uuid
        if self.handle != '':
            return self.handle
        return None

    def __eq__(self, other):
        if self.uuid == '' and other.uuid == '' and self.handle == '' and other.handle == '':
            raise ValueError('Can not compare objects without a uuid or handle.')
        return (self.uuid == other.uuid) if self.uuid != '' else (self.handle == other.handle)

    def __str__(self):
        return f'DSpace object with the uuid {self.uuid}:\n\t' + '\n\t'.join(str(self.metadata).split('\n'))

    def to_dict(self) -> dict:
        """
            Converts the current item object to a dictionary object containing all available metadata.
        """
        return {'uuid': self.uuid, 'handle': self.handle, 'name': self.name, 'metadata': self.metadata}

    def get_metadata_values(self, tag: str) -> list | None:
        """
        Retrieves the metadata values of a specific tag as a list.

        :param tag: The metadata tag: prefix.element.qualifier
        :return: The values as a list or None, if the tag doesn't exist.
        """
        m = self.metadata.get(tag)
        return [v.value for v in m] if m else None

    def add_statistic_report(self, report: dict | list[dict] | None):
        """
        Adds a new report or list of reports as a dict object to the DSpaceObject

        :param report: The report(s) to add.
        """
        if report is None:
            return
        report = [report] if isinstance(report, dict) else report
        for r in report:
            for k in r.keys():
                if k not in self.statistic_reports.keys():
                    self.statistic_reports[k] = r[k]
                else:
                    if isinstance(r[k], dict):
                        self.statistic_reports[k] = (self.statistic_reports[k] +
                                                     [r[k]]) if isinstance(self.statistic_reports[k],
                                                                           list) else [self.statistic_reports[k], r[k]]
                    else:
                        self.statistic_reports[k] = r[k]

    def has_statistics(self) -> bool:
        """
        Checks if statistic reports are available for this object.

        :return: True if there is at least one report.
        """
        return len(self.statistic_reports.keys()) > 0



import warnings
from io import BytesIO
from PIL import Image

from dspyce.bitstreams import Bitstream


class IIIFBitstream(Bitstream):
    """
        A class for managing iiif-specific content files in the saf-packages.
    """

    iiif: dict[str, str | int]
    """
        A dictionary containing the IIIF-specific information. The keys must be: 'label', 'toc', 'w', 'h'
    """

    def __init__(self, name: str, path: str, bundle: any = None, uuid: str = None, primary: bool = False):
        """
        Creates a new IIIF-Bitstream object.

        :param name: The name of the bitstream.
        :param path: The path, where the file is currently stored.
        :param bundle: The bundle, where the bitstream should be placed in. The default is ORIGINAL.
        :param uuid: The uuid of the bitstream if existing.
        :param primary: Primary is used to specify the primary bitstream.
        """
        super().__init__(name, path, bundle, uuid, primary)
        self.iiif = {}

    def __str__(self):
        """
        Provides all information about the DSpace IIIF-Content file.

        :return: A SAF-ready information string which can be used for the content-file.
        """
        export_name = super().__str__()
        if len(self.iiif.keys()) > 0:
            export_name += (f'\tiiif-label:{self.iiif["label"]}'
                            f'\tiiif-toc:{self.iiif["toc"]}'
                            f'\tiiif-width:{self.iiif["w"]}'
                            f'\tiiif-height:{self.iiif["h"]}')
        else:
            warnings.warn('You are about to generate information of IIIF-specific DSpace bitstream, without providing'
                          'IIIF-specific information. Are you sure, you want to do this?')
        return export_name

    def add_iiif(self, label: str, toc: str, w: int = 0):
        """
            Add  IIIF-information for the bitstream.

            :param label: is the label that will be used for the image in the viewer.
            :param toc: is the label that will be used for a table of contents entry in the viewer.
            :param w: is the image width to reduce it. Default 0
        """
        img = Image.open(BytesIO(self.get_bitstream_file()))
        width, height = img.size
        if w != 0 and w < width:
            scale = int(width/w)
            super().file = img.reduce(scale).tobytes()
        img.close()
        self.iiif = {'label': label, 'toc': toc, 'w': width, 'h': height}



# saf/saf_write.py
"""
Python module for creating saf packages to prepare DSpace Item-imports or updates. Two functions are imported by the saf
package:
1. create_saf_package() -> Creates a single saf package based on a given Item.
2. saf_packages() -> Creates multiple saf packages in a given path based on the given Item information.
"""
import logging
import os
from bs4 import BeautifulSoup
from bs4.formatter import XMLFormatter

from dspyce.Item import Item
from dspyce.Relation import Relation
from dspyce.metadata import MetaData


LOG_LEVEL = logging.INFO
"""A global log_level for logging. Is used by all sub-modules as a default value for the log-level."""

LOG_FILE: str | None = None
"""
A global log-file for logging. Is used by all sub-modules as default. If "NONE" all logs will be printed into the
console.
"""

LOG_FORMAT: str = '%(asctime)s - %(levelname)s: %(message)s'
"""A global log-format setting, wich is used by all sub-modules."""


def export_relations(relations: list[Relation]) -> str:
    """
        Creates a list of relationships separated by line-breaks. It can be used to create the relationship-file in a
        saf-package.

        :param relations: A list of objects of the class "Relation"
        :return: The line-break separated list of relationships as a string.
    """
    relation_strings = list(map(lambda x: ':'.join(str(x).split(':')[1:]), relations))
    return '\n'.join([r.replace(':', ' ') for r in relation_strings])


def save_text_file(path: str, file_name: str, content: str):
    """
    Save a text into path using encoding *utf-8*. Raises an exception if the file already exists.

    :param path: The path where to store the text-file.
    :param file_name: The name of the new file.
    :param content: The content of the text-file.
    :raises FileExistsError: If the file already exits in path.
    """
    if file_name in os.listdir(path):
        raise FileExistsError(f'The file {file_name} already exists in path {path}.')
    with open(f'{path}/{file_name}', 'w', encoding='utf8') as f:
        f.write(content)
        logging.debug(f'Wrote file "{file_name}" into "{path}"')


def create_bitstreams(item: Item, save_path: str):
    """
       Creates the need bitstream-files in the archive-directory based on the path information.

        :param item: The item to create the bitstreams for.
        :param save_path: The path, where the bitstream shall be saved.
   """
    logging.basicConfig(level=LOG_LEVEL, filename=LOG_FILE, encoding='utf8',
                        format=LOG_FORMAT)
    contents = []
    bundles = item.get_bundles()
    for bundle in bundles:
        for b in bundle.bitstreams:
            if b.bundle.name != bundle.name:
                logging.error(f'The name of the bundle({bundle.name}) containing the bitstream({b.file_name}) does not'
                              f'match the name of the bundle registered in the bitstream object: {b.bundle.name}')
            contents.append(str(b))

            b.save_bitstream(save_path)
            logging.debug(f'Successfully created file {b.file_name}')
    if len(contents) > 0:
        save_text_file(save_path, 'contents', '\n'.join(contents))


def export_schemas(metadata: MetaData) -> dict:
    """
    Creates the content of the xml files metadata_[prefix].xml based on the given metadata.

    :param metadata: The metadata object containing all metadata fields.
    :return: The content of the xml file for this schema prefix in a dictionary.
    """
    class SAFFormatter(XMLFormatter):
        def attributes(self, tag):
            for k, l in tag.attrs.items():
                yield k, l

    xml_schemas = {}
    for prefix in metadata.get_schemas():
        schema_xml = BeautifulSoup(features='xml', preserve_whitespace_tags=["dcvalue"])
        file_name = f'metadata_{prefix}.xml'
        if prefix == 'dc':
            file_name = 'dublin_core.xml'
            schema_xml.append(schema_xml.new_tag('dublin_core'))
        else:
            schema_xml.append(schema_xml.new_tag('dublin_core', schema=prefix))
        metadata_dict = metadata.get_by_schema(prefix)
        for m in metadata_dict.keys():
            tag = m.split('.')
            element, qualifier = tag[1], tag[2] if len(tag) > 2 else None
            for v in metadata_dict[m]:
                mv_field = schema_xml.new_tag("dcvalue", element=element)
                if qualifier is not None:
                    mv_field["qualifier"] = qualifier
                if v.language is not None and v.language != '':
                    mv_field["language"] = v.language
                mv_field.string = v.value
                schema_xml.contents[0].append(mv_field)
        xml_schemas[file_name] = schema_xml.prettify(formatter=SAFFormatter())
    return xml_schemas


def create_saf_package(item: Item, element_id: int, path: str, overwrite: bool = False):
    """
    Creates a saf package folder for an item object.

    :param item: The Item to create the package of.
    :param element_id: An id added to the directory name, aka item_<element_id>
    :param path: The path where to store all package files.
    :param overwrite: If true, it overwrites the currently existing files.
    """

    logging.basicConfig(level=LOG_LEVEL, filename=LOG_FILE, encoding='utf8',
                        format=LOG_FORMAT)

    path += '/' if len(path) > 0 and path[-1] != '/' else ('./' if len(path) == 0 else '')
    if 'archive_directory' not in os.listdir(path):
        os.mkdir(path + 'archive_directory')
        logging.info(f'Created directory *archive_directory* in "{path}"')
    path += 'archive_directory/'
    # Now we check, if the element already exists and deleted it, if overwrite is set to true.
    if f'item_{element_id}' in os.listdir(path):
        logging.debug(f'The Item "{element_id}" already exists in "{path}".')
        if not overwrite:
            logging.error(f'The item with the id {element_id} already exists in "{path}"!')
            raise FileExistsError(f'The item with the element_id "{element_id}" exists already'
                                  f'in "{path}"')
        else:
            logging.debug(f'Deleted old item "{element_id}" in "{path}"')
            for file in os.listdir(f'{path}/item_{element_id}'):
                os.remove(f'{path}/item_{element_id}/{file}')
            os.rmdir(f'{path}/item_{element_id}')

    os.mkdir(f'{path}item_{element_id}')
    logging.info(f'Created directory for item with the ID {element_id}')
    path += f'item_{element_id}/'

    xml_files = export_schemas(item.metadata)
    for file_name in xml_files.keys():
        save_text_file(path, file_name, str(xml_files[file_name]))
        logging.debug(f'Wrote %s to item folder item_%s' % (file_name, element_id))

    if len(item.relations) > 0:
        save_text_file(path, 'relations', export_relations(item.relations))
        logging.debug(f'Created relations file for item {element_id}')
    if item.handle != '':
        save_text_file(path, 'handle', item.handle)
        logging.debug(f'Created handle file for item {element_id}')
    if len(item.collections) > 0:
        save_text_file(path, 'collections', '\n'.join([c.get_identifier() for c in item.collections]))
        logging.debug(f'Created collections file for item {element_id}')
    create_bitstreams(item, save_path=path)


def saf_packages(items: list[Item], path: str, overwrite: bool = False):
    """
    Creates a list of saf packages based on a given item list and writes them down in the filesystem.

    :param items: The list of items to create the packages from.
    :param path: The path in the filesystem where to store the packages.
    :param overwrite: If true, it overwrites the currently existing files.
    """
    logging.basicConfig(level=LOG_LEVEL, filename=LOG_FILE, encoding='utf8',
                        format=LOG_FORMAT)
    n = 0
    show_progress = len(items) >= 1000
    logging.info(f'Start process! Processing {len(items)}')
    for item in items:
        create_saf_package(item, n, path, overwrite)
        if show_progress and n % 100 == 0:
            logging.info(f'\tCreated {n} saf packages.')
        n += 1
    logging.info(f'Finished process! Created {len(items)} saf packages.')



import re


class MetaDataValue:
    """
    This class represents a metadata field for a DSpace object. Containing information about the schema, element,
    qualifier and value.
    """
    language: str | None
    """The language of the metadata value. For example 'en' for English."""
    value: str | int | float | bool
    """The actual metadata value."""

    def __init__(self, value: str | int | float | bool, language: str = None):
        """
        Creates a new MetaDataValue object.

        :param value: The value stored in the metadata field.
        :param language: Optional language parameter for a metadata field.
        """
        self.value = value
        self.language = language if language != '' else None

    def __eq__(self, other):
        """
        Checks if to Metadata objects ar equal.

        :param other: The metadata object to compare to
        :return: True if the two objects are equal, False otherwise
        :raises TypeError: If the parameter other is not of Type Metadata.
        """
        if not isinstance(other, MetaDataValue):
            raise TypeError(f'Can not use = between type(MetadataValue) and type({type(other)})')
        other: MetaDataValue
        return (isinstance(self.value, type(other.value)) and
                self.value == other.value and self.language == other.language)

    def __str__(self):
        """
        Creates a string representation of this object.
        """
        return (f'{self.language}:\t' if self.language else '') + str(self.value)

    def set_value(self, value):
        """
        Adds a value to the metadata field, if it already has a value, the old value will be overwritten.
        :param value: The value to add.
        """
        self.value = value

    def __iter__(self):
        """
        Returns a dictionary representation of this object.
        """
        yield 'value', self.value
        if self.language is not None:
            yield 'language', self.language


class MetaData(dict):
    """
    A dict of metadata fields using the following format: "<tag>": list(MetaDataValue)
    """

    @staticmethod
    def is_valid_tag(tag) -> bool:
        """
        Checks if the given tag is valid using RegEx.

        :param tag: The tag to check.
        :return: True if the tag is valid, False otherwise.
        """
        if re.search(r'^[a-zA-Z0-9\-]+\.[a-zA-Z0-9\-]+(\.[a-zA-Z0-9\-]+)?$', tag):
            return True
        else:
            return False

    def __setitem__(self, key: str, value: MetaDataValue | list[MetaDataValue]):
        """
        Adds a value to the metadata field defined by key. If the metadata field already exists, the value will be
        appended. If value is a list of MetaDataValues, the values will be replaced with the new value list.

        :param key: The key of the metadata field to add (must be in the correct format).
        :param value: The value to add. If this is of type list, the new list of MetadataValues will replace the old one
        :raises TypeError: If the parameter value is not of type MetaDataValue.
        :raises KeyError: If the parameter key has not a valid format.
        """
        if not isinstance(value, MetaDataValue) and not isinstance(value, list):
            raise TypeError(f'The value must be of type MetaDataValue, but found {type(value)}')
        if not MetaData.is_valid_tag(key):
            raise KeyError(f'The key "{key}" is not a valid metadata key.')
        if isinstance(value, list):
            for v in value:
                if not isinstance(v, MetaDataValue):
                    raise TypeError(f'All values must be of type MetaDataValue, but found {type(value)}')
            super().__setitem__(key, value)
        else:
            if key not in self.keys():
                super().__setitem__(key, [value])
            else:
                super().__getitem__(key).append(value)

    def get_schemas(self) -> set[str]:
        """
        Creates a list of used schemas in the metadata list.

        :return: A list of schema names a strings.
        """
        return {k.split('.')[0] for k in self.keys()}

    def __getitem__(self, item) -> list[MetaDataValue]:
        return super().__getitem__(item)

    def get(self, key, default=None) -> list[MetaDataValue]:
        return super().get(key, default)

    def __str__(self):
        """
        Creates a string representation of the Metadata object.
        """
        return '\n'.join([f'{k}:\n' + '\n'.join([f'\t{v}' for v in self.get(k)]) for k in self.keys()])

    def get_by_schema(self, schema: str):
        """
        Returns a sub-dictionary only including metadata fields from the given schema.

        :param schema: The schema name.
        :return: A sub-dictionary of the given schema.
        :raises KeyError: If the given schema does not have any metadata fields.
        """
        if schema not in self.get_schemas():
            raise KeyError(f'The schema "{schema}" is not used.')
        return MetaData({k: self.__getitem__(k) for k in filter(lambda x: x.split('.')[0] == schema, self.keys())})
